/** @file console.c
 *  @brief A console driver.
 *
 *  This driver implements the console. The console
 *  maintains its own cursor position, term color,
 *  and cursor visibility, independent of the hardware
 *  cursor, so that the cursor can be freely moved
 *  despite the actual visibility status.
 *
 *  @author Lee Salzman (lsalzman)
 *  @bug No known bugs.
 */

#include <console.h>
#include <string.h>
#include <x86/pio.h>
#include <x86/proc_reg.h>

#define CONSOLE_CELL(x,y) ((char *) CONSOLE_MEM_BASE + 2 * ((y) * CONSOLE_WIDTH + (x)))
 
/* the current term color */
static int consoleColor = FGND_LGRAY | BGND_BLACK,
/* the current cursor column */
           cursorX = 0,
/* the current cursor row */
           cursorY = 0,
/* the cursor visibility status */
           cursorHidden = 1;

/** @brief Erases a part of the console.
 *
 *  Overwrites the specified screen region with the
 *  current console color.
 *
 *  @param row the row of upper left corner of the region
 *  @param col the column of the upper left corner of the region
 *  @param w the width of the region
 *  @param h the height of the region
 *  @param overwrite whether or not to overwrite the color
 *
 *  @return Void
 */

void
console_erase (int row, int col, int w, int h, int overwrite)
{
    /* the starting position in memory of the current row */
    char * start = CONSOLE_CELL (col, row);

    while (h > 0)
    {
        /* the current position in the row */
        char * pos = start,
        /* the end of the row */
             * end = start + 2 * w;

        while (pos < end)
        {
            /* write in blank spaces with the current console color */
            pos [0] = ' ';

	    if (overwrite)
              pos [1] = consoleColor;

            pos += 2;
        }

        /* move to the next row */
        start += 2 * CONSOLE_WIDTH;
        -- h;
    }
}
        
/** @brief Scrolls the console.
 *
 *  Scroll the console down a specified number of lines
 *  and clear the new space.
 *
 *  @param lines the number of lines to scroll
 *
 *  @return Void
 */

void
console_scroll (unsigned lines)
{
    /* don't allow the console to scroll any farther than the cursor */
    if (lines > cursorY)
      lines = cursorY;

    /* move the console contents up a number of lines */
    memcpy (CONSOLE_CELL (0, 0),
            CONSOLE_CELL (0, lines),
            CONSOLE_CELL (0, CONSOLE_HEIGHT) - CONSOLE_CELL (0, lines));
    
    /* clear the new space that was scrolled onto the console */
    console_erase (CONSOLE_HEIGHT - lines, 0, CONSOLE_WIDTH, lines, 1);

    /* reposition the cursor relative to the scroll*/
    set_cursor (cursorY - lines, cursorX);
}

/** @brief Backs up the cursor.
 *
 *  Move the cursor backward a specified distance
 *  and clear the characters along the way.
 *
 *  @param distance the number of characters to backspace
 *
 *  @return Void
 */

void
console_backspace (unsigned distance)
{
    /* clip the backspace distance to go no farther than the first column */
    if (distance > cursorX)
      distance = cursorX;

    /* clear the backspaced characters */
    console_erase (cursorY, cursorX - distance, distance, 1, 0);

    /* reposition the cursor */
    set_cursor (cursorY, cursorX - distance);
}

int
putbyte( char ch )
{
    disable_interrupts ();

    switch (ch)
    {
    case '\r':
        /* move the cursor to the beginning of the row */
        set_cursor (cursorY, 0);
        break;

    case '\n':
        /* if there's no more vertical space, scroll down */
        if (cursorY + 1 == CONSOLE_HEIGHT)
          console_scroll (1);
        /* move the cursor to the beginning of the next row */
        set_cursor (cursorY + 1, 0);
        break;

    case '\b':
        /* backspace one character */
        console_backspace (1);
        break;

    default:
        draw_char (cursorY, cursorX, ch, consoleColor);
            
        /* if there is room, move the cursor forward one column */
        if (cursorX + 1 < CONSOLE_WIDTH)
          set_cursor (cursorY, cursorX + 1);
        else
        {
            /* if there's no more vertical space, scroll down */
            if (cursorY + 1 == CONSOLE_HEIGHT)
              console_scroll (1);
            /* move the cursor to the beginning of the next line */
            set_cursor (cursorY + 1, 0);
        }
    }
    
    enable_interrupts ();
    
    return ch;
}

void 
putbytes( const char *s, int len )
{
    if (s == NULL)
      return;

    /* print each character one at a time */
    while (len > 0)
    {
        putbyte (* s);

        ++ s;
        -- len;
    }
}

void
set_term_color( int color )
{
    /* colors must be an 8 bit integer */
    if (color < 0 || color > 0xFF)
      return;

    consoleColor = color;
}

void
get_term_color( int * color )
{
    if (color == NULL)
      return;
    
    * color = consoleColor;
}

/** @brief Set the CRTC cursor position.
 *
 *  Outputs the position to the CRTC cursor register.
 *
 *  @param pos 16-bit cursor position
 *
 *  @return Void
 */

static void
crtc_set_cursor (unsigned short pos)
{
    outb (CRTC_IDX_REG, CRTC_CURSOR_LSB_IDX);
    outb (CRTC_DATA_REG, pos & 0xFF);
    outb (CRTC_IDX_REG, CRTC_CURSOR_MSB_IDX);
    outb (CRTC_DATA_REG, pos >> 8);
}

void
set_cursor( int row, int col )
{
    /* ensure the row and column are within screen space */
    if (row < 0 || row >= CONSOLE_HEIGHT ||
        col < 0 || col >= CONSOLE_WIDTH)
      return;
    
    cursorY = row;
    cursorX = col;

    /* don't update the hardware cursor if hidden */
    if (cursorHidden)
      return;

    /* output the new position to the hardware cursor */
    crtc_set_cursor (row * CONSOLE_WIDTH + col);
}

void
get_cursor( int * row, int * col )
{
    if (row == NULL || col == NULL)
      return;

    * row = cursorY;
    * col = cursorX;
}

void
hide_cursor()
{
    /* don't hide the cursor if it's already hidden */
    if (cursorHidden)
      return;

    /* move the cursor out of bounds */
    crtc_set_cursor (0xFFFF);

    cursorHidden = 1;
}

void
show_cursor()
{
    /* don't show the cursor if it's already shown */
    if (! cursorHidden)
      return;

    /* position the hardware cursor back on the screen */
    crtc_set_cursor (cursorY * CONSOLE_WIDTH + cursorX);

    cursorHidden = 0;
}

void 
clear_console()
{
    /* clear the screen space */
    console_erase (0, 0, CONSOLE_WIDTH, CONSOLE_HEIGHT, 1);
}

void
draw_char( int row, int col, int ch, int color )
{
    char * pos;
    
    /* ensure the row and column are within screen space 
     * and the color and character are valid 8 bit integers
     * the character must not move the cursor
     */
    if (row < 0 || row >= CONSOLE_HEIGHT ||
        col < 0 || col >= CONSOLE_WIDTH ||
        ch < 0 || ch > 0xFF ||
        ch == '\r' || ch == '\b' || ch == '\n' ||
        color < 0 || color > 0xFF)
      return;

    /* write out the character and color to the position */ 
    pos = CONSOLE_CELL (col, row);

    pos [0] = ch;
    pos [1] = (char) color;
}

char
get_char( int row, int col )
{
    /* ensure row and column are within screen space */
    if (row < 0 || row >= CONSOLE_HEIGHT ||
        col < 0 || col >= CONSOLE_WIDTH)
      return '\0';
 
    return * CONSOLE_CELL (col, row);
}
