/** @file frame.c
 *  @brief Provides the implementantion of physical page frame allocation.
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */

#include <frame.h>
#include <malloc.h>
#include <stdlib.h>
#include <x86/proc_reg.h>

static FrameAllocator frameAllocator;

void
frame_allocator_reserve (FrameAddress start, FrameAddress end)
{
	/* Allocate individual pages that are available within the specified region. */
	while (start < end)
	{
		frame_allocate_within (0, start, start);
		
		++ start;
	}
}
	
static void
frame_allocator_fill (FrameAddress start, FrameAddress end)
{
	/* Fill the frame allocator with pages in a given physical memory region */
	FrameAddress alignedAddress = start, 
		address; 
	FrameAlignment maxAlign, alignment;
	Frame * frame;
	List * freeList;

	/* find the largest alignment that can be allocated within the region */
	for (maxAlign = 0; maxAlign <= FRAME_ALIGNMENT_MAXIMUM; ++ maxAlign)
	{
		if ((1 << maxAlign) > end - start + 1)
		  break;
	}

	-- maxAlign;
	if ((1 << maxAlign) == end - start + 1 && alignedAddress % (1 << maxAlign))
	  -- maxAlign;

	/* find the aligned address of the first maximally aligned flexible frame in the region */
	if (alignedAddress % (1 << maxAlign))
	  alignedAddress += (1 << maxAlign) - (alignedAddress % (1 << maxAlign));

	/* insert flexible frames greeding, from the largest alignment down to the smallest, into the free lists starting at the aligned address */
	address = alignedAddress;
	frame = & frameAllocator.frames [address - frameAllocator.frameStart];
	alignment = maxAlign;

	for (freeList = & frameAllocator.freeLists [alignment]; freeList >= frameAllocator.freeLists; -- freeList)
	{
		while (address + (1 << alignment) <= end)
		{
			list_insert_front (freeList, frame);
			
			frame -> alignment = alignment;

			address += 1 << alignment;
			frame += 1 << alignment;
		}

		-- alignment;
	}

	/* start from the aligned address and work backwards, greedily inserting frames from largest alignment down to smallest that fits */
        address = alignedAddress;
        frame = & frameAllocator.frames [address - frameAllocator.frameStart];
        alignment = maxAlign - 1;

        for (freeList = & frameAllocator.freeLists [alignment]; freeList >= frameAllocator.freeLists; -- freeList)
        {
                while (address - (1 << alignment) >= start)
                {
			address -= 1 << alignment;
			frame -= 1 << alignment;

                        list_insert_back (freeList, frame);

                        frame -> alignment = alignment;
                }

                -- alignment;
        }
}
		
int
frame_allocator_initialize (FrameAddress start, FrameAddress end)
{
	Frame * frame;
	FrameAlignment alignment;

	/* set the extent of the frame allocator region and allocate frame meta-data */
	frameAllocator.framesAvailable = end - start + 1;
	frameAllocator.frameStart = start;
	frameAllocator.frameEnd = end;
	frameAllocator.frames = (Frame *) calloc (end - start + 1, sizeof (Frame));

	if (frameAllocator.frames == NULL)
	  return -1;

	/* clear the free lists */
	for (alignment = 0; alignment <= FRAME_ALIGNMENT_MAXIMUM; ++ alignment)
	{
		list_clear (& frameAllocator.freeLists [alignment]);
	}

	/* initialize the addresses and alignments for the frame meta-data */
	for (frame = frameAllocator.frames; frame <= & frameAllocator.frames [end - start]; ++ frame)
	{
		frame -> address = start + (frame - frameAllocator.frames);
		frame -> alignment = ~0;
	}

	/* fill the free lists with all pages in the region */
	frame_allocator_fill (start, end);
 
	return 0;
}

static Frame *
frame_allocator_split (Frame * frame, FrameAlignment request, Frame * start, Frame * end)
{
    FrameAlignment alignment = frame -> alignment - 1;
    List * freeList = & frameAllocator.freeLists [alignment];

    /* attempt to split a page until it meets the requested alignment */

    if (! frame -> allocated)
      list_remove ((ListIterator) frame);

    /* find the frames buddy for its current alignment, split it off, and then insert it into the free list */
    while (freeList >= & frameAllocator.freeLists [request])
    {
	Frame * buddy = frame + (1 << alignment);

	frame -> alignment = alignment;
	buddy -> alignment = alignment;

	/* ensure the chosen frame of the two buddies is within the memory region */
	if (buddy < end)
	{
		list_insert_back (freeList, frame);

		frame = buddy;
	}
	else
	{
		list_insert_back (freeList, buddy);
	}

	-- freeList;
	-- alignment;
    }

    frame -> allocated = 1;

    return frame;
}

static Frame *
frame_allocator_attempt_split (FrameAlignment request, Frame * start, Frame * end)
{
    List * freeList;

    /* look for a frame to split down to a requested alignment, if any, from frames bigger than the requested alignment */
    for (freeList = & frameAllocator.freeLists [request + 1]; freeList <= & frameAllocator.freeLists [FRAME_ALIGNMENT_MAXIMUM]; ++ freeList)
    {
	ListIterator position;

        if (list_empty (freeList))
          continue;

	for (position = list_begin (freeList); position != list_end (freeList); position = list_next (position))
	{
		Frame * frame = (Frame *) position;

		/* only accept splits that either fall within the region or have sub-frames falling within the region */
		if (frame >= start && frame <= end)
		  return frame_allocator_split (frame, request, start, end);

		if (frame < start && frame + (1 << frame -> alignment) - (1 << request) >= start)
		  return frame_allocator_split (frame, request, start, end);
	}
    }

    return NULL;
}

static Frame *
frame_allocator_coalesce (Frame * frame, FrameAlignment request, Frame * start, Frame * end)
{
    FrameAlignment alignment = frame -> alignment;

    /* coalesce a frame with its buddies up to the requested alignment */

    list_remove ((ListIterator) frame);

    /* remove the buddy and group it with the current frame, while the buddy is available or until the request is met */
    while (alignment < request)
    {
        Frame * buddy = & frameAllocator.frames [(frame -> address ^ (1 << alignment)) - frameAllocator.frameStart];

	if (buddy < frameAllocator.frames ||
	    buddy > & frameAllocator.frames [frameAllocator.frameEnd - frameAllocator.frameStart] ||
	    buddy -> alignment != alignment ||
	    buddy -> allocated)
	  break;

        list_remove ((ListIterator) buddy);

        if (buddy < frame)
          frame = buddy;

	++ alignment;
    }

    frame -> alignment = alignment;

    /* check if the frame is within the specified memory region */
    if (alignment == request && frame >= start && frame <= end)
    {
        frame -> allocated = 1;

        return frame;
    }

    /* otherwise, put the coalesced frame back and fail */
    list_insert_front (& frameAllocator.freeLists [alignment], frame);

    return NULL;
}

static Frame *
frame_allocator_attempt_coalesce (FrameAlignment request, Frame * start, Frame * end)
{
    List * freeList;
    FrameAlignment alignment = request - 1;

    /* look for frames to coalesce into a new frame of the requested alignment, working down from the requested alignment */
    for (freeList = & frameAllocator.freeLists [alignment]; freeList >= frameAllocator.freeLists; -- freeList)
    {
        ListIterator position = list_begin (freeList);

	/* scan through the free list looking for a suitable coalescing candidate */
        while (position != list_end (freeList))
        {
            Frame * frame = (Frame *) position,
                  * buddy;

            position = list_next (position);

	    /* if the candidate is outside the range, try the next one */
            if (frame < start || frame >= end + (1 << request))
              continue;

	    /* check if the frame's buddy is available for coalescing and if is the right size */
            buddy = & frameAllocator.frames [(frame -> address ^ (1 << alignment)) - frameAllocator.frameStart];

            if (buddy >= frameAllocator.frames &&
                buddy <= & frameAllocator.frames [frameAllocator.frameEnd - frameAllocator.frameStart] &&
                buddy -> alignment == alignment &&
                buddy -> allocated == 0)
	    {
                Frame * frame;

                if (position == (ListIterator) buddy)
                  position = list_next (position);
	
                frame = frame_allocator_coalesce (frame, request, start, end);

		/* if the coalescing succeeded, then return the coalesced frame */ 
                if (frame != NULL)
                  return frame;
	    }
        }

        -- alignment;
    }

    return NULL;
}

Frame *
frame_allocate_within (FrameAlignment alignment, FrameAddress start, FrameAddress end)
{
	List * freeList;
	ListIterator position;
	Frame * frame, * startFrame, * endFrame;

	/* sanity check the alignment and region to make sure the request could even possibly be fulfilled */
	if (alignment > FRAME_ALIGNMENT_MAXIMUM ||
	    start > end ||
	    start > frameAllocator.frameEnd ||
	    end < frameAllocator.frameStart)
	  return NULL;

	disable_interrupts ();

	/* determine the acceptable allocation boundary frames */
	startFrame = & frameAllocator.frames [start - frameAllocator.frameStart];
	endFrame = & frameAllocator.frames [end - frameAllocator.frameStart];
	freeList = & frameAllocator.freeLists [alignment];

	/* look for any frames in the requested alignment's free list that are in the acceptable range */
	for (position = list_begin (freeList); position != list_end (freeList); position = list_next (position))
	{
		frame = (Frame *) position;

		/* if the frame is in range, allocate it and return */
		if (frame >= startFrame && frame <= endFrame)
		{
			list_remove (position);

			frame -> allocated = 1;

			frameAllocator.framesAvailable -= 1 << alignment;

			enable_interrupts ();

			return frame;
		}
	}	
	
	/* the free list does not contain any acceptable pages
	 * so try to coalesce any smaller frames into an acceptable frame.
	 * if coalescing failed, try to split any larger frames to the an acceptable
	 * frame of the requested alignment
	 */
	frame = frame_allocator_attempt_coalesce (alignment, startFrame, endFrame);
	if (frame == NULL)
	  frame = frame_allocator_attempt_split (alignment, startFrame, endFrame);

	if (frame != NULL)
	  frameAllocator.framesAvailable -= 1 << alignment;

	enable_interrupts ();

	return frame;
}

Frame *
frame_allocate (FrameAlignment alignment)
{
	/* allocate a frame from anywhere within the frame allocator's manage physical memory */
	return frame_allocate_within (alignment, frameAllocator.frameStart, frameAllocator.frameEnd);
}

void
frame_free (Frame * frame)
{
	/* mark the frame as free and insert it back into the free list */
	disable_interrupts ();

	frame -> allocated = 0;

	list_insert_front (& frameAllocator.freeLists [frame -> alignment], frame);

	frameAllocator.framesAvailable += 1 << frame -> alignment;

	enable_interrupts ();
}

unsigned int
frame_availability (void)
{
	return frameAllocator.framesAvailable;
}
		

