/** @file id.c
 *  @brief Implemention of the IDTable tasks
 *
 *  This file provides the implementation for the ID tables for threads 
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */

#include <id.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>

int
id_table_create (IdTable * table, unsigned int size)
{
  /*set the values to create a new IDTable */
  table -> nextId = 1;
  table -> entryCount = size;
  table -> entries = (IdEntry *) calloc (size, sizeof (IdEntry));  /*allocate memory for  entries */
  
  if (table -> entries == NULL)     /*Return -1 if table->entries is NULL */
    return -1;
  
  return 0;
}


void
id_table_destroy (IdTable * table)
{
  free (table -> entries);  /*free the memory used for the IDTable*/
}

static IdEntry *
id_table_find_entry (IdTable * table, int hash, int key)
{
	IdEntry * entry;

	hash %= table -> entryCount;  /*find the location of a specific entry*/

	/*look for the correct entry in the table, linearly probing starting at the hash */
	for (entry = & table -> entries [hash]; entry < & table -> entries [table -> entryCount]; ++ entry)
	{
		if (entry -> id == key)
		  return entry;                  
	}

	/*finally check the rest of the entries from the beginning up to the hash */
        for (entry = table -> entries; entry < & table -> entries [hash]; ++ entry)
        {
                if (entry -> id == key)
                  return entry;
        }
 
	/* no entry found */
	return NULL;
}

void *
id_table_find (IdTable * table, int id)
{

  /*find a specific entry in the table */
  IdEntry * entry = id_table_find_entry (table, id, id);
  
  if (entry != NULL)
    return entry -> data;
  
  return NULL;
}

void *  
id_table_remove (IdTable * table, int id)
{       
  IdEntry * entry = id_table_find_entry (table, id, id); /*find the location of a specific entry */
  void * data = NULL;
 
  /*if found; remove the entry */
  if (entry != NULL)
    {
      data = entry -> data;
      entry -> id = 0;
      entry -> data = NULL;
    }
  
  /*return the entry data */
  return data;
}

int
id_table_insert (IdTable * table, void * data)
{
  /*looks for a place to insert the entry */
  IdEntry * entry = id_table_find_entry (table, table -> nextId, 0);
  
  /*If it cannot find a place; double the table size to insert the entry */
  if (entry == NULL)
    {
      unsigned int oldSize = table -> entryCount;
      IdEntry * oldEntries = table -> entries,
	* newEntries = (IdEntry *) calloc (oldSize * 2 + 1, sizeof (IdEntry)),
	* oldEntry;
      if (newEntries == NULL)
	return -1;
      
      table -> entryCount = oldSize * 2 + 1;
      table -> entries = newEntries;
      
      /*Reinstall the entries into the grown hash table */
      for (oldEntry = oldEntries; oldEntry < & oldEntries [oldSize]; ++ oldEntry)
	{
	  if (! oldEntry -> id)
	    continue;
	  
	  entry = id_table_find_entry (table, oldEntry -> id, 0);
	  entry -> id = oldEntry -> id;
	  entry -> data = oldEntry -> data;
	}
      
      /*Free the old hash table */
      free (oldEntries);
      
      entry = id_table_find_entry (table, table -> nextId, 0);
    }
  
  /* assign the data the next id + 1 */
  entry -> id = table -> nextId ++;
  entry -> data = data;
  
  return entry -> id;
}

