/** @file frame.h
 *  @brief Interface for the frame allocator
 *
 *  Interface for physical page frame allocation.
 *
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_FRAME_H__
#define __KERN_FRAME_H__

#include <page.h>
#include <list.h>

/* the size of page frames in bytes */
#define FRAME_SIZE 		PAGE_SIZE
/* the number of significant bits in a frame address */
#define FRAME_ADDRESS_BITS 	(32 - PAGE_SHIFT)
/* the maximal number of bits that may be used to represent a frame's alignment */
#define FRAME_ALIGNMENT_BITS 	(PAGE_SHIFT - 1)
/* the maximum alignment a flexible page may be, which is the number of power of two sizes possible */
#define FRAME_ALIGNMENT_MAXIMUM	(FRAME_ADDRESS_BITS - 1)

typedef unsigned int FrameAddress;
typedef unsigned int FrameAlignment;

/* structure representing a flexible page */
typedef struct
{
	/* list node for the list of free physical frames of a given alignment */
	ListNode frameList;
	/* the frame's address, with its non-significant bits shifted off */
	unsigned int address : FRAME_ADDRESS_BITS;
	/* the frame's alignment, which is also it's size */
	unsigned int alignment : FRAME_ALIGNMENT_BITS;
	/* whether or not the frame has been allocated */ 
	unsigned int allocated : 1;
} Frame;

/* the actual memory address of a frame, which is the frame's index multiplied by the byte size of a frame */
#define FRAME_ADDRESS(frame)	((frame) -> address * FRAME_SIZE)

/* a frame allocator managing a region of physical address space */
typedef struct
{
	/* the number of frames available to allocate */ 
	unsigned int framesAvailable;
	/* the starting frame address of the managed memory */
	FrameAddress frameStart;
	/* the last useable frame address of the managed memory */
	FrameAddress frameEnd;
	/* the frame meta-data for all managed frames */
	Frame * frames;
	/* free lists of frames of given alignments up to the maximum alignment */
	List freeLists [FRAME_ALIGNMENT_MAXIMUM + 1];
} FrameAllocator;

/** @brief Initializes the frame allocator
 *
 * Initialize the frame allocator to manage a specified region of physical memory.
 *
 * @param start the frame address of the start of the managed region
 * @param end the frame address of the last frame of the managed region
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int frame_allocator_initialize (FrameAddress start, FrameAddress end);
/** @brief Reserve frame from the allocator
 *
 * Reserves page frames from the allocator so that they may not be used.
 *
 * @param start the frame address of the start of the reserved frames
 * @param end the frame address of the last frame of the reserved frames
 *
 * @return Void
 */ 
extern void frame_allocator_reserve (FrameAddress start, FrameAddress end);

/** @brief Allocate a frame within a memory region
 *
 * Allocates a frame of the specified alignment within a given region of the physical address space.
 *
 * @param alignment the alignment of the frame
 * @param start the frame address of the start of the region
 * @param end the frame address of the last acceptable firame of the region
 *
 * @return The allocated frame or NULL on failure.
 */
extern Frame * frame_allocate_within (FrameAlignment alignment, FrameAddress start, FrameAddress end);
/** @brief Allocate a frame
 *
 * Allocate a frame from anywhere within the managed physical memory.
 *
 * @param alignment the alignment of the frame
 *
 * @return The allocated frame or NULL on failure.
 */
extern Frame * frame_allocate (FrameAlignment alignment);
/** @brief Free a frame
 *
 * Return the frame to the free list
 * 
 * @param frame the frame to free
 *
 * @return Void
 */
extern void frame_free (Frame * frame);

/** @brief The number of available frames for allocation
 *
 * Return the number of frames available for allocation, if allocated at the least alignment possible.
 *
 * @return The number of frames.
 */
extern unsigned int frame_availability (void);

#endif /* __KERN_FRAME_H__ */


