/** @file id.h
 *  @brief Id-mapping facilities.
 *
 *  Provides the interface for id mapping facilities for threads.
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */


#ifndef __KERN_ID_H__
#define __KERN_ID_H__

/* a mapping from an individual id to assocated data */
typedef struct IdEntry
{
	/* the thread id */
	int id;
	/* the associated data */
	void * data;
} IdEntry;

/* a hash table of ids mapping to thread data */
typedef struct
{
	/* the next id to be assigned */
	int nextId;
	/* the number of entries in the table */
	unsigned int entryCount;
	/* the entries of the table */
	IdEntry * entries;
} IdTable;

/** @brief Creates an id table.
 *
 * Creates a new id table of the specified size.
 *
 * @param table the location of the table
 * @param size the starting size of tha table
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int id_table_create (IdTable * table, unsigned int size);
/** @brief Destroys an id table.
 *
 * Frees any memory used by the id table.
 *
 * @param table the table to destroy
 *
 * @return Void
 */
extern void id_table_destroy (IdTable * table);

/** @brief Allocate an id.
 *
 * Insert a thread into the id table with a new id.
 *
 * @param table the table to insert into
 * @param data the thread data
 * 
 * @return the allocated id on success
 *         -1 on failure
 */ 
extern int id_table_insert (IdTable * table, void * data);
/** @brief Remove an id from the table.
 *
 * Removes an id from the table.
 *
 * @param table the table to remove from
 * @param id the id to remove
 *
 * @return the data that was at the entry if present
 *         NULL if not present
 */
extern void * id_table_remove (IdTable * table, int id);

/** @brief Find an id in the table.
 *
 * Returns the data associated with a thread id.
 *
 * @param table the table to find in
 * @param id the id to find
 *
 * @return the data associated with the id on success
 *         NULL on failure
 */
extern void * id_table_find (IdTable * table, int id);

#endif /* __KERN_ID_H__ */

