/** @file interrupt.h
 *  @brief Interrupt handler interface.
 *
 *  Provides the interface for interrupt handling.
 *
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_INTERRUPT_H__
#define __KERN_INTERRUPT_H__

#include <list.h>
#include <register.h>

/* the size of the stack used for handling interrupts */
#define INTERRUPT_STACK_SIZE (16 * 1024)

/* the size in bytes of each interrupt stub */
#define INTERRUPT_STUB_SIZE 16

/* the code for an individual interrupt stub */
typedef char InterruptStub [16];

/* the array of interrupt stubs to be installed into the IDT */
extern InterruptStub interrupt_stubs [];

/* the interrupt vector of the first exception */
#define INTERRUPT_VECTOR_EXCEPTIONS 0
/* the interrupt vector of the first hardware interrupt */
#define INTERRUPT_VECTOR_INTERRUPTS 32
/* the interrupt vector of the first software trap */
#define INTERRUPT_VECTOR_TRAPS 48
/* the number of the maximum interrupt */
#define INTERRUPT_VECTOR_MAXIMUM 99

typedef unsigned int InterruptVector;

typedef void (* InterruptHandlerFunction) (InterruptVector, Registers *);

/* a handler for an interrupt */
typedef struct
{
	/* the position of the handler in the handler list */
	ListNode interruptHandlerList;
	/* a function to be invoked when the interrupt is signaled */
	InterruptHandlerFunction handlerFunction;
	/* the vector of the interrupt to wait for */
	InterruptVector interruptVector;
} InterruptHandler;

typedef unsigned int ExceptionVector;
typedef unsigned int ExceptionData;

typedef void (* ExceptionHandlerFunction) (ExceptionVector, ExceptionData, Registers *);

/* a handler for an exception interrupt */
typedef struct
{
	/* the position of the handler in the handler list */
	ListNode exceptionHandlerList;
	/* a function to be invoked when the exception is raised */
	ExceptionHandlerFunction handlerFunction;
	/* the vector of the exception to wait for */
	ExceptionVector exceptionVector;
} ExceptionHandler;

/** @brief Initialize the interrupt dispatcher.
 *
 * Initialize the IDT with the interrupt stubs and the handler lists.
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int interrupt_dispatcher_initialize (void);

/** @brief Create a new interrupt handler.
 * 
 * Create a new handler for the specified interrupt vector, with a given handler function.
 *
 * @param vector the vector of the interrupt to wait for
 * @param handler the handler for the interrupt
 *
 * @return an interrupt handler on success
 *         NULL on failure
 */ 
extern InterruptHandler * interrupt_handler_create (InterruptVector vector, InterruptHandlerFunction handler);
/** @brief Destroy an interrupt handler.
 *
 * Removes an interrupt handler and frees any memory it used.
 *
 * @param handler the handler to destroy
 * 
 * @return Void
 */
extern void interrupt_handler_destroy (InterruptHandler * handler);
/** @brief Create a new exception handler.
 * 
 * Create a new handler for the specified exception vector, with a given handler function.
 *
 * @param vector the vector of the exception to wait for
 * @param handler the handler for the exception
 *
 * @return an exception handler on success
 *         NULL on failure
 */
extern ExceptionHandler * exception_handler_create (ExceptionVector vector, ExceptionHandlerFunction handler);
/** @brief Destroy an exception handler.
 *
 * Removes an exception handler and frees any memory it used.
 *
 * @param handler the handler to destroy
 * 
 * @return Void
 */
extern void exception_handler_destroy (ExceptionHandler * handler);

/** @brief Wait for any interrupts.
 *
 * Halt the kernel until an interrupt occurs.
 *
 * @return Void
 */
extern void interrupt_wait (void);

#endif /* __KERN_INTERRUPT_H__ */

