/** @file io.h
 *  @brief Interface for the Input Output system calls
 *
 *  This file provides the interface for i/o system calls. It handles the system calls pretaining the user interface.
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_IO_H__
#define __KERN_IO_H__



#include <x86/seg.h>      /* for install_user_segs() */
#include <x86/pic.h>      /* for pic_init() */
#include <x86/base_irq.h> /* for base_irq_master/slave */
#include <x86/irq_list.h>
#include <x86/proc_reg.h>
#include <keyhelp.h>
#include <interrupt.h>
#include <interrupts.h>
#include <list.h>
#include <register.h>
#include <syscall_int.h>
#include <keyhelp.h>
#include <vm.h>
#include <thread.h>
#include <scheduler.h>
#include <console.h>
#include <stdio.h>
#include <string.h>

/*The handle to the keyboard interrupt in the interrupt vector */
#define KEYBOARD_INT INTERRUPT_VECTOR_INTERRUPTS+IRQ_KEYBOARD

/*The buffer size for the input buffer */
#define KEYBOARD_BUFFER_SIZE 4096

/*Struct to get arguments for readline */
struct rl{
  /*get the number of chars to read */
  int len;
  /*Points to there buffer where to put the chars once they've been read */
  char* buf;
};

/*Struct to hold the setcursor arguments */
struct sc{
  /*holds the row number */
  int row;
  /*holds the col number */
  int col;
};

/*Struct to hold the getcursor arguments */
struct gc{
  /*Pointer to the row */
  int *row;
  /*Pointer to the col */
  int *col;
};


/** @brief Initialize the I/O interrupts
 *
 *  Installs the I/O interrups
 *
 *  @return returns 0 on success and 1 on failure                    
 */
int io_initialize();


/** @brief Initializes the keyboard interrupt
 *
 *  Its the keyboard interrupt handler 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
void keyboard_handler(InterruptVector vector, Registers *registers);



/** @brief Gets a new char from the queue
 *
 *   Its the getchar interrupt handler 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
void getchar_handler(InterruptVector vector, Registers *registers);


/** @brief Gets a newline from the queue
 *
 *   Its ths readline interrupt handler
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
void readline_handler(InterruptVector vector, Registers *registers);


/** @brief Changes the color of the console
 *
 *   Its the settermcolor interrupt handler 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
void termcolor_handler(InterruptVector vector, Registers *registers);


/** @brief Gets the cursor position on the console
 *
 *   Its the get_cursor_pos interrupt handler 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
void getcursor_handler(InterruptVector vector, Registers *registers);


/** @brief Sets the cursor position on the console
 *
 *   Its the get_cursor_pos interrupt handler 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
void setcursor_handler(InterruptVector vector, Registers *registers);

/** @brief Prints a line to the console
 *
 *   Its the print interrupt handler 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
void print_handler(InterruptVector vector, Registers *registers);


/** @brief Processess the scan from the character queue
 *
 *   Removes a character from the queue
 *
 *  @return returns the value of the scancode
 */
int process_scode(void);


/** @brief The handler function for getchar 
 *
 *  Handles the getchar syscall and thread waiting on a resourse that have been woken up
 *
 *  @param thread   The thread being woken up
 *  @param data     Paramaters to the function
 *
 *  @return Void 
 */
void getchar_wake_handler(Thread* thread, void *data);


/** @brief The handler function for readline
 *
 *  Handles the readline syscall and thread waiting on a resourse that have been woken up
 *
 *  @param thread   The thread being woken up
 *  @param data     Paramaters to the function
 *
 *  @return Void 
 */
void readline_wake_handler(Thread* thread, void *data);

/** @brief Inserts new character into the queue
 *
 *   Adds new characters to the back of the queue
 *
 *  @param c            The character to be added to the queue
 *
 *  @return Void 
 */
void insert(int c);


#endif
