/** @file list.h
 *  @brief Inline list structure facilities.
 *
 *  Provides an interface for inline lists that store the nodes directly in data.
 *  Lists maintain a sentinel node and are doubly linked.
 *
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_LIST_H__
#define __KERN_LIST_H__

/* a list node */
typedef struct ListNode
{
    /* the next node in the list */
    struct ListNode * next;
    /* the previous node in the list */
    struct ListNode * previous;
} ListNode;

/* an abstract iterator for the list */
typedef ListNode * ListIterator;

/* a list structure which contains the sentinel node
 * the sentinel's next is the first item in the list
 * the sentinel's previous is the last item in the list
 * if no item are in the list the sentinel points to itself
 */
typedef struct
{   
   /* the sentinel node of the list */
    ListNode sentinel;
} List;

#define list_begin(list)        ((list) -> sentinel.next)
#define list_end(list)          (& (list) -> sentinel)
#define list_next(iterator)     ((iterator) -> next)
#define list_previous(iterator) ((iterator) -> previous)

#define list_front(list)        ((void *) list_begin (list))
#define list_back(list)         ((void *) list_previous (list_end (list)))

#define list_empty(list)        (list_begin (list) == list_end (list))

#define list_insert_front(list,value)   (list_insert (list_begin (list), (value)))
#define list_insert_back(list,value)    (list_insert (list_end (list), (value)))

#define list_remove_front(list) (list_remove (list_begin (list)))
#define list_remove_back(list)  (list_remove (list_previous (list_end (list))))

#define LIST_ELEMENT(node,type,field) \
	((type *) ((char *) (node) - (unsigned int) & ((type *) 0) -> field))

/** @brief Clear the list.
 *
 * Clears the list to be empty.
 *
 * @param list the list to clear
 *
 * @return Void
 */
extern void list_clear (List * list);
/** @brief Insert into a list.
 * 
 * Inserts a new value before the specified position in a list.
 *
 * @param position iterator for the list position
 * @param value the value to insert into the list
 *
 * @return the iterator of the inserted item
 */
extern ListIterator list_insert (ListIterator position, void *value);
/** @brief Remove from a list.
 *
 * Remove the specified position from a list.
 *
 * @param position iterator for the list position
 * 
 * @return the value at the position that was removed
 */
extern void * list_remove (ListIterator position);
/** @brief Count list elements.
 *
 * Return the number of elements in a list.
 *
 * @param list the list to count
 *
 * @return the number of elements in the list
 */ 
extern unsigned int list_size (List * list);

#endif /* __KERN_LIST_H__ */
 
