/** @file scheduler.h
 *  @brief Scheduler interface.
 *
 *  Provides the interface for scheduling threads and for scheduling resources.
 *
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_SCHEDULER_H__
#define __KERN_SCHEDULER_H__

#include <list.h>
#include <register.h>

struct Thread;

/* a scheduling resource for threads to wait upon */
typedef struct
{
	/* the list of waiting threads */
	List waitingThreads;
} Resource;

/* a thread scheduler for managing ready/running threads */
typedef struct
{
	/* the currently running thread */
	struct Thread * currentThread; 
	/* any threads ready to run immediately */
	List readyThreads;
} Scheduler;

/** @brief Initialize the scheduler.
 *
 * Initiliazes the thread queues for the scheduler.
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int scheduler_initialize (void);

/** @brief Tick the scheduler.
 *
 * Implements the scheduler tick, which decrements the quantum of the currently running thread.
 *
 * @return Void
 */
extern void scheduler_tick (void);
/** @brief Context switch to the next thread to run.
 *
 * Context switches into the next thread that should be run, storing the interrupt registers as necessary.
 *
 * @param registers the interrupt registers
 *
 * @return Void
 */
extern void scheduler_context_switch (Registers * registers);
/** @brief Jump into first ready thread.
 *
 * Moves into the first ready thread and activates scheduling.
 *
 * @return Void
 */
extern void scheduler_jump (void);

/** @brief Schedule a thread to be ready.
 *
 * Inserts a thread into the scheduler ready list.
 *
 * @param thread the thread to make ready
 *
 * @return Void
 */
extern void scheduler_schedule (struct Thread * thread);
/** @brief Unschedule a ready or running thread.
 *
 * Removes a thread from the ready list, or removes the currently running thread if that is the specified
 * thread.
 *
 * @param thread the thread to unschedule
 *
 * @return Void
 */
extern void scheduler_unschedule (struct Thread * thread);
/** @brief Yield to a thread.
 *
 * Pushes the currently running thread back onto the ready list
 * and optionally moves the specified thread to the front of the ready list,
 * so that it will be run upon return from interrupts.
 *
 * @param thread the thread to yield to, or NULL if any thread may be yielded to
 *
 * @return Void
 */
extern void scheduler_yield (struct Thread * thread);

/** @brief Create a resource.
 *
 * Initialize the waiting thread list of the resource.
 * 
 * @param resource the location of the resource
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int resource_create (Resource * resource);
/** @brief Destroy a resource.
 *
 * Free any memory associated with the resource.
 *
 * @param resource the resource to destroy
 * 
 * @return Void
 */
extern void resource_destroy (Resource * resource);

/** @brief Wait on a resource.
 *
 * Inserts a thread onto the back of a resource's wait queue, installing a wait handler to be 
 * invoked when the resource signals the thread to wake up.
 *
 * @param resource resource to wait on
 * @param thread the waiting thread
 * @param waitingHandler the handler to be called when the thread is woken up
 * @param waitingData argument to the wake handler
 *
 * @return Void
 */
extern void resource_wait (Resource * resource, struct Thread * thread, void (* waitingHandler) (struct Thread *, void *), void * waitingData);
/** @brief Rewait on a resource.
 *
 * Inserts a thread at the front of the resource's wait queue, reusing any wake handlers the thread had
 * previously specified the last time it waited.
 *
 * @param resource resource to wait on
 * @param thread the waiting thread
 *
 * @return Void
 */ 
extern void resource_rewait (Resource * resource, struct Thread * thread);
/** @brief Signal threads waiting on resource.
 *
 * Wake up a specified amount of threads waiting on the resource and invoke their wake handlers.
 *
 * @param resource the resource to signal
 * @param amount the amount of threads to attempt to wake, or -1 if no limit
 *
 * @return the number of threads awakened
 */
extern int resource_wake (Resource * resource, int amount);

#endif /* __KERN_SCHEDULER_H__ */

