/** @file task.h
 *  @brief Task manager interface.
 *
 *  Provides the interface for task management.
 *
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_TASK_H__
#define __KERN_TASK_H__

#include <list.h>
#include <frame.h>
#include <scheduler.h>

/* Max. size of the name of a file */
#define EXECNAME_LIMIT		256
/* Max size of the parameters sent to the task */
#define EXECVEC_ARGV_LIMIT	256
/* Max size of the buffer */
#define EXECVEC_BUFFER_LIMIT	4096

struct Thread;

typedef struct Task
{
        /*Task list for the tasks */
	ListNode taskList;
        /*The parent task */
	struct Task * parentTask;
        /*List of teh child tasks */
	List childTasks;	
        /*The page the page directory is on */
	Frame * pageTable;
        /*The list of the page tables */
	List pageTables;
        /* Task address space */
	List memoryRegions;
	struct Thread * mainThread;
        /*List of threads running */
        List threads;
        /*Resource to monitor the exit status of threads */
	Resource exitResource;
       /*exitId for the task */
	int exitId;
       /*The status of the task */
	int exitStatus;
} Task;


/** @brief Initializes the task manager
 *
 *  runs the init task 
 *
 *  @return returns 0 on success and -1 on failure                    
 */
extern int task_manager_initialize (void);


/** @brief Gets the currently running task
 *
 *  returns the task of the currently running thread 
 *
 *  @return NULL if no thread is running or the task of the currently running thread 
 */
extern Task * task_current (void);

/** @brief Creates a new task
 *
 *  Creates a new task and inserts it to the back of the list
 *
 *  @param parent the parent of the task to be created 
 * 
 *  @return NULL if task couldn't be created; the task otherwise
 */
extern Task * task_create (Task * parent);


/** @brief Destroys a task 
 *
 *  Kills a task that was in the list of tasks
 *
 *  @param task the task to be killed
 * 
 *  @return Void
 */
extern void task_destroy (Task * task);


/** @brief Forks a task
 *
 * Forks a new task that is a copy of the current task
 *
 * @return the forked task on success
 *         NULL on failure
 */
extern Task * task_fork (void);


extern int task_exec (Task * task, char * execname, char ** execvec);

/** @brief Free resources for an exiting task
 * 
 * This should be called once the last thread of a task has exited to
 * free any resources associated with the task and wake any threads waiting
 * for it to complete.
 *
 * @param task the task to exit
 *
 * @return Void
 */
extern void task_exit (Task * task);
 
#endif /* __KERN_TASK_H__ */

