/** @file thread.h
 *  @brief Thread manager interface.
 *
 *  Provides the interface for managing threads.
 *
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_THREAD_H__
#define __KERN_THREAD_H__

#include <list.h>
#include <task.h>
#include <register.h>
#include <timer.h>

/* the run states a thread may be in */
typedef enum
{
	/* unscheduled, it is not waiting on any resource nor is ready to run/running */
	THREAD_STATE_UNSCHEDULED = 0,
	/* ready to run */
	THREAD_STATE_READY,
	/* the thread is currently running */
	THREAD_STATE_RUNNING,
	/* the thread is waiting on a resource */
	THREAD_STATE_WAITING
} ThreadState;

/* the initial quantum is 1/10 of a second, or 100 ms */
#define THREAD_QUANTUM_INITIAL (TIMER_HZ / 10)

typedef unsigned int ThreadQuantum;

/* the thread control block */
typedef struct Thread
{
	/* the position the thread holds in its scheduling list */
	ListNode schedulerList;
	/* the position of the threads in its task's thread list */
	ListNode threadList;
	/* the id of the thread */
	int id;
	/* the task containing the thread */
	Task * task; 
	/* the current run state of the thread */
	ThreadState state;
	/* the quantum remaining for the thread */
	ThreadQuantum quantum;
	/* the saved registers for the thread to be restored upon running */
	Registers registers;
	/* the wake handler for the thread when it is woken from sleeping on a resource */
	void (* waitingHandler) (struct Thread *, void *);
	/* data for the wake handler */
	void * waitingData;
} Thread;

/** @brief Initialze thread manager.
 *
 * Initialize the thread manager's id tables.
 *
 * @return 0 on success
 *        -1 on failure
 */ 
extern int thread_manager_initialize (void);

/** @brief Currently running thread.
 *
 * Return the currently running thread, if any.
 *
 * @return the running thread on success
 *         NULL on failure
 */
extern Thread * thread_current (void);

/** @brief Create a new thread.
 *
 * Create a new thread in a specified task.
 * 
 * @param task the task to create a thread in
 * 
 * @return the new thread on success
 *         NULL on failure
 */
extern Thread * thread_create (Task * task);
/** @brief Destroy a thread.
 *
 * Free the thread's memory resources and remove it from its task.
 *
 * @param thread the thread to destroy
 *
 * @return Void
 */
extern void thread_destroy (Thread * thread);

/** @brief Fork a new thread.
 *
 * Fork the current thread to run in a specified task.
 *
 * @param task the task to fork a new thread into
 *
 * @return the forked thread on success
 *         NULL on failure
 */
extern Thread * thread_fork (Task * task);
/** @brief Exit from a thread.
 *
 * Exit from a thread, setting the exit status of the task and exiting
 * from the host task if it is the last.
 *
 * @param thread the thread to exit from
 * @param exitStatus the exit status of the thread
 *
 * @return Void
 */
extern void thread_exit (Thread * thread, int exitStatus);

/** @brief Find a thread by id.
 *
 * Find the thread associated with a given thread id.
 *
 * @param id the id of the thread
 *
 * @return the found thread on success
 *         NULL on failure
 */
extern Thread * thread_find (int id);

#endif /* __KERN_THREAD_H__ */

