/** @file timer.h
 *  @brief Timer facilities interface.
 *
 *  Provides the interface for simple deadline timers.
 * 
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_TIMER_H__
#define __KERN_TIMER_H__

#include <list.h>
#include <scheduler.h>

#define TIMER_HZ 100

/* a deadline timer that signals a resource upon expiration */
typedef struct
{
	/* the list of timers waiting to expire */
	ListNode timerList;
	/* the tick deadline of the timer */
	unsigned int deadline;
	/* resource for threads waiting on the timer */
	Resource sleepResource;
} Timer;

/** @brief Initialize timer facilities.
 * 
 * Initialize the timer facilities and start up the timer.
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int timer_initialize (void);

/** @brief The number of ticks since boot.
 *
 * Return the number of ticks elapsed since initializing the timer.
 *
 * @return the number of ticks
 */
extern unsigned int get_ticks (void);

/** @brief Create a new timer.
 *
 * Create a new timer to expire a certain number of ticks from now.
 * The timer must be activated after threads have been suspended on the resource before it will
 * wait for expiration.
 *
 * @param delta the number of ticks that should elapse from now till expiration
 *
 * @return a timer on success
 *         NULL on failure
 */
extern Timer * timer_create (unsigned int delta);
/** @brief Activate a timer.
 *
 * Install the timer so that it will expire at its deadline.
 *
 * @param timer the timer to activate
 *
 * @return Void
 */
extern void timer_activate (Timer * timer);
/** @brief Destroy a timer.
 *
 * Free any memory resources associated with a timer.
 *
 * @param timer the timer to destroy
 *
 * @return Void
 */
extern void timer_destroy (Timer * timer);
 
#endif /* __KERN_TIMER_H__ */


