/** @file vm.h
 *  @brief Virtual memory management interface.
 *
 *  Provides the interface for managing memory regions and the virtual address space.
 *
 *  @bug No known bugs.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __KERN_VIRTUAL_MEMORY_H__
#define __KERN_VIRTUAL_MEMORY_H__

#include <list.h>
#include <task.h>
#include <register.h>

/* the starting address of the kernel heap */
#define VM_ADDRESS_KERNEL_HEAP			0x00000000
/* the size of the kernel heap */
#define VM_ADDRESS_KERNEL_HEAP_LIMIT		0x01000000
/* the starting address of the user space memory */
#define VM_ADDRESS_USER_SPACE			0x01000000
/* the size of the user space memory */
#define VM_ADDRESS_USER_SPACE_LIMIT		0xFE000000
/* the address of the first temporary mapping space, which is one superpage in size */
#define VM_ADDRESS_TEMPORARY_MAPPING_1		0xFF000000
/* the address of the second temporary mapping space, which is one superpage in size */
#define VM_ADDRESS_TEMPORARY_MAPPING_2		0xFF400000
/* the address of the first temporary page directory mapping */
#define VM_ADDRESS_TEMPORARY_MAPPING_1_PT0	0xFF800000
/* the address of the second temporary page directory mapping */
#define VM_ADDRESS_TEMPORARY_MAPPING_2_PT0	0xFF801000
/* the starting address of the reserved kernel stack space */
#define VM_ADDRESS_KERNEL_STACK			0xFF802000
/* the size of the reserved kernel stack space */
#define VM_ADDRESS_KERNEL_STACK_LIMIT		0x003FE000
/* the address of the self-mapped page table structures, which is one superpage in size */
#define VM_ADDRESS_PAGE_TABLE			0xFFC00000

/* memory protection flag for readable memory regions */
#define VM_PROTECTION_READ		(0 << 1)
/* memory protection flag for writeable memory regions */
#define VM_PROTECTION_WRITE		(1 << 1)
/* memory protection flag for supervisor memory regions */
#define VM_PROTECTION_SUPERVISOR	(0 << 2)
/* memory protection flag for user memory regions */
#define VM_PROTECTION_USER		(1 << 2)
/* mask of memory protection flags */
#define VM_PROTECTION_MASK		(3 << 1)

/* the address of the page directory */
#define VM_PT0                          ((PageTableEntry *) (VM_ADDRESS_PAGE_TABLE + ((VM_ADDRESS_PAGE_TABLE >> VM_PT0_SHIFT) << VM_PT1_SHIFT)))
/* the address mask for page directory entries */
#define VM_PT0_MASK			0xFFC00000
/* the offset of the address mask for page directory entries */
#define VM_PT0_SHIFT			22
/* the number of entries in a page directory */
#define VM_PT0_SIZE			1024

/* the address of the page table at the specified index in the page directory */
#define VM_PT1(index)                   ((PageTableEntry *) (VM_ADDRESS_PAGE_TABLE + (index) * sizeof (PageTableEntry) * VM_PT1_SIZE))
/* the address mask for page table entries */
#define VM_PT1_MASK                     0x003FF000
/* the offset of the address mask for page table entries */
#define VM_PT1_SHIFT                    12
/* the number of entries in a page table */
#define VM_PT1_SIZE			1024

/* the page table entry associated with a given address */
#define VM_PTE(address)			(VM_PT1((address) >> VM_PT0_SHIFT) [((address) & VM_PT1_MASK) >> VM_PT1_SHIFT])

/* the page table entry flag for marking a PTE as present */
#define VM_PTE_PRESENT                 	(1 << 0)
/* the page table entry flag for marking a PTE as global */
#define VM_PTE_GLOBAL                  	(1 << 8)

/* the acceptable distance from a page fault to the current top of the stack that is tolerated for growing the stack */
#define VM_STACK_TOLERANCE		(1 << VM_PT0_SHIFT)
	
typedef unsigned int PageTableEntry;

typedef unsigned long int MemoryAddress;

struct MemoryRegion;

typedef int (* MemoryFaultHandler) (struct MemoryRegion *, MemoryAddress, Registers *);

/* a memory region in a task's address space */ 
typedef struct MemoryRegion
{
	/* the position in a task's memory region list */
	ListNode memoryRegionList;	
	/* the memory protection of the region */
	unsigned int protection;
	/* if the memory region grows downward, the amount of distance from the page fault to the base of the region it will tolerate */
	int growsDownward;
	/* the address of the starting page in the task's address space */
	MemoryAddress addressStart;
	/* the address of the ending page in the task's address space */
	MemoryAddress addressEnd;
	/* the number of frames allocated to the region */
	unsigned int frameCount;
	/* the list of frames allocated to the region */
	List frames;
	/* a handler to be called when a page fault occurs within the region or its tolerance */
	MemoryFaultHandler faultHandler;
} MemoryRegion;

/** @brief Initialize virtual memory.
 *
 * Initializes the virtual memory system with the specified init task.
 *
 * @param task the init task
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int vm_initialize (Task * task);

/** @brief Validates that a memory region is mapped.
 *
 * Checks if the specified region of the current task's address space is mapped as user memory with the specified protection.
 *
 * @param start the address of beginning of the region
 * @param end the address of the last frame of the region
 * @param protection the protection flags to check for
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int vm_validate (MemoryAddress start, MemoryAddress end, unsigned int protection);
/** @brief Validates that a memory region is mapped.
 *
 * Checks if the specified region of the current task's address space is mapped as user readable, where the memory region
 * is a string containing elements of a specified size with an unknown extent, ending at a 0 element.
 *
 * @param start the address of beginning of the region
 * @param size the element size of the string memory region
 *
 * @return the length of the string 0 on success
 *         -1 on failure
 */
extern int vm_validate_string (MemoryAddress start, unsigned int size);

/** @brief Context switch into a task's address space.
 *
 * Installs the page directory of the specified task as the current page directory.
 *
 * @param task the task to context switch into
 *
 * @return Void
 */
extern void vm_context_switch (Task * task);

/** @brief Map a memory region into a task's address space.
 *
 * Map a memory region into a task's address space, allocating page tables as necessary.
 *
 * @param task the task to map into
 * @param region the memory region to map
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int vm_map (Task * task, MemoryRegion * region);
/** @brief Unmap a memory region from a task's address space.
 *
 * Unset all the page table entries associated with a memory region in the task's address space.
 *
 * @param task the task to unmap from
 * @param region the region to unmap
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int vm_unmap (Task * task, MemoryRegion * region);


/** @brief Forks a new task from the current task.
 *
 * Forks a new address space that is a copy of the current task into the specified task.
 *
 * @param task the task to fork the address space into
 * 
 * @return 0 on success
 *         -1 on failure
 */
extern int vm_fork (Task * task);
/** @brief Copies data into a specified task's address space.
 *
 * Copies data from the current address space into a target address space by utilizing the temporary mapping space.
 *
 * @param task the task to copy data into
 * @param dst the memory address in the target task's address space to copy into
 * @param src the data from the current address space to copy
 * @param size the amount of data to copy
 *
 * @return 0 on success
 *         -1 on failure
 */
extern int vm_copy (Task * task, MemoryAddress dst, char * src, unsigned int size);

/* memory region fault handler for creating stacks */
extern int stack_fault_handler (MemoryRegion *, MemoryAddress, Registers *);

/** @brief Creates a new memory region in a task's address space.
 * 
 * Creates a new memory region in the tasks address space, for a given protection and fault behavior.
 *
 * @param task the task to create in
 * @param protection the mapping protection of the region
 * @param start the address of the first frame of the region
 * @param end the address of the last frame of the region
 * @param handler the page fault handler for the region
 *
 * @return the memory region on success
 *         NULL on failure
 */
extern MemoryRegion * memory_region_create (Task * task, unsigned int protection, MemoryAddress start, MemoryAddress end, MemoryFaultHandler handler);
/** @brief Destroys a memory region.
 *
 * Free frames associated with a memory region.
 * 
 * @param region the region to destroy
 * 
 * @return Void
 */
extern void memory_region_destroy (MemoryRegion * region);
/** @brief Finds a memory region by address.
 *
 * Find a memory region in a task by an address that falls within the region.
 *
 * @param task the task to find the memory region in
 * @param address the address falling within a region to search for
 *
 * @return the region containing the address on success
 *         NULL on failure
 */ 
extern MemoryRegion * memory_region_find (Task * task, MemoryAddress address);

/* macro that automatically calculates the end address for vm_validate */
#define VM_VALIDATE(ptr,type,elems)	(vm_validate ((MemoryAddress) (ptr), (MemoryAddress) (ptr) + sizeof (type) * (elems) - 1, VM_PROTECTION_READ))
/* macro that automatically calculates the end address for vm_validate, also checks for writeable memory */
#define VM_VALIDATE_WRITE(ptr,type,elems)     (vm_validate ((MemoryAddress) (ptr), (MemoryAddress) (ptr) + sizeof (type) * (elems) - 1, VM_PROTECTION_READ | VM_PROTECTION_WRITE))
/* convenient macro for vm_validate_string that passes in the right type size */
#define VM_VALIDATE_STRING(ptr,type)	(vm_validate_string ((MemoryAddress) (ptr), sizeof (type))) 

#endif /* __KERN_VIRTUAL_MEMORY_H__ */

