/** @file interrupt.c
 *  @brief Provides the implementantion for installing interrupts 
 *
 *  Takes care of installing a handling interrupts and exceptions 
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */

#include <interrupt.h>
#include <register.h>
#include <list.h>
#include <vm.h>
#include <scheduler.h>
#include <interrupts.h>
#include <console.h>
#include <x86/gate_init.h>
#include <x86/proc_reg.h>
#include <x86/seg.h>
#include <x86/pic.h>
#include <assert.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <thread.h>

/*List of interrupt handlers */
static List interruptHandlers [INTERRUPT_VECTOR_MAXIMUM + 1];

int
interrupt_dispatcher_initialize (void)
{
	char gate [sizeof (struct x86_gate)];
	InterruptVector vector;

	memset (gate, 0, sizeof (gate));

	for (vector = 0; vector <= INTERRUPT_VECTOR_MAXIMUM; ++ vector)
	{
		unsigned int stub = (unsigned int) & interrupt_stubs [vector];

		list_clear (& interruptHandlers [vector]);

		/* 32 bit trap gate, segment present, kernel privilege level */
		gate [5] = ACC_TRAP_GATE | ACC_P | (vector >= INTERRUPT_VECTOR_TRAPS ? ACC_PL_U : ACC_PL_K);
		/* low 16 bits of the interrupt stub */
		* (unsigned short *) (gate + 0) = stub & 0xFFFF;
		/* kernel code segment */
		* (unsigned short *) (gate + 2) = KERNEL_CS_SEGSEL;
		/* high 16 bits of the interrupt stub */
		* (unsigned short *) (gate + 6) = stub >> 16;
   
		/* install trap gate at idt entry */
		memcpy ((char *) sidt () + vector * sizeof (struct x86_gate), 
			gate, 
			sizeof (struct x86_gate));
	}

	//set_esp0 (malloc (4096));
	set_esp0 ((char *) VM_ADDRESS_KERNEL_STACK + VM_ADDRESS_KERNEL_STACK_LIMIT);

	return 0;
}
	
InterruptHandler *
interrupt_handler_create (InterruptVector vector, InterruptHandlerFunction function)
{
	InterruptHandler * handler;

	/*If size is Max. then return Null */
	if (vector > INTERRUPT_VECTOR_MAXIMUM)
	  return NULL;

	/*Malloc space for a new handler */
	handler  = (InterruptHandler *) malloc (sizeof (InterruptHandler));
	if (handler == NULL)
	  return NULL;

	/*setup handler */
	handler -> handlerFunction = function;
	handler -> interruptVector = vector;

	disable_interrupts ();

	/*Insert the handler into the list */
	list_insert_back (& interruptHandlers [vector], handler);

	enable_interrupts ();
	
	return handler;
}

void
interrupt_handler_destroy (InterruptHandler * handler)
{
	disable_interrupts ();

	/*remove the handler */
	list_remove ((ListIterator) handler);

	enable_interrupts ();
	/*free the handler */
	free (handler);
}

ExceptionHandler *
exception_handler_create (ExceptionVector vector, ExceptionHandlerFunction function)
{
  /*installs the exception handler */
  assert (sizeof (ExceptionHandler) == sizeof (InterruptHandler));
  
  return (ExceptionHandler *) interrupt_handler_create ((InterruptVector) vector,
					(InterruptHandlerFunction) function);
}

void
exception_handler_destroy (ExceptionHandler * handler)
{
  /*destroy the interrupt handler */
  interrupt_handler_destroy ((InterruptHandler *) handler);
}

void
exception_dispatch (ExceptionVector vector, ExceptionData data, Registers * registers)
{
	static char buffer [512];
	ListIterator position;

	if (list_empty (& interruptHandlers [vector]))
	{
	  /*If exception handler is not present; kill the thread for a graceful return */
	    Thread * thread = thread_current ();

	    thread = thread_current ();
 	    if (thread != NULL)
            {
                putbytes (buffer, snprintf (buffer, sizeof (buffer),
                        "Thread %d killed by exception %d\n"
                        "  EIP: %x\n",
                        thread -> id, vector, registers -> EIP));

                thread_exit (thread, -2);
            }
	}
	else /*find the position of the exception handler */
	for (position = list_begin (& interruptHandlers [vector]);
	     position != list_end (& interruptHandlers [vector]);
	     position = list_next (position))
	{
	    ExceptionHandler * handler = (ExceptionHandler *) position;
	      
	    (* handler -> handlerFunction) (vector, data, registers);
	}
	/*if returning from an exception handler; context switch to the next runnlable thread */
	if (registers -> CS == USER_CS_SEGSEL)
	  scheduler_context_switch (registers);
}

void
interrupt_dispatch (InterruptVector vector, Registers * registers)
{
        ListIterator position; 

	/*find the correct interrupt handler */
        for (position = list_begin (& interruptHandlers [vector]);
             position != list_end (& interruptHandlers [vector]);
             position = list_next (position))
        {
	  /*Invokes the interrupt handler*/
                InterruptHandler * handler = (InterruptHandler *) position;
                (* handler -> handlerFunction) (vector, registers);
        }
	
	if (vector >= INTERRUPT_VECTOR_INTERRUPTS && vector < INTERRUPT_VECTOR_TRAPS)
	  pic_ack (vector - INTERRUPT_VECTOR_INTERRUPTS); /*for the hardware interrupt; clear hardware */

	/*if returning from an exception handler; context switch to the next runnlable thread */
	if (registers -> CS == USER_CS_SEGSEL)
	  scheduler_context_switch (registers);
}

void
interrupt_wait (void)
{
  /*halt */
  asm volatile ("hlt");
}
 
