/** @file io.c
 *  @brief Implemention of the i/o syscalls 
 *
 *  This file provides the implementation for i/o system calls.
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */
#include <io.h>

static int back;                                    /* points to the back of the circular queue*/
static int front;                                   /* points to the front of the circular queue*/
static int curr_size;                               /* keeps count of the number of char in the queue */
static char buffer [KEYBOARD_BUFFER_SIZE];          /* circular queue */
static char rl_buffer [KEYBOARD_BUFFER_SIZE];       /* readline buffer to be used as a temp buffer */
static int rl_index;                                /* index into the readline buffer */
static int rl_thread;                               /* the thread that owns the readline buffer */
static Resource keyboardResource;                   /* Resource on which readline & getchar have to wait */

int io_initialize(){
  front=0;                                          /* Initialize the front of queue */
  back=-1;                                          /* Initialize the back of queue */
  curr_size=0;                                      /* Set current size of buffer to 0 */
  rl_thread=-1;                                     /* Set thrad id to -1 */

  resource_create(&keyboardResource);

  if (interrupt_handler_create(KEYBOARD_INT, &keyboard_handler) < 0 ||               /*Installing the keyboard interrupt */
      interrupt_handler_create(GETCHAR_INT, getchar_handler) < 0 ||                  /*Installing the getchar interrupt */        
      interrupt_handler_create(READLINE_INT, readline_handler) < 0 ||                /*Installing the readline interrupt */
      interrupt_handler_create(SET_TERM_COLOR_INT, termcolor_handler) < 0 ||         /*Installing the set_term_color interrupt */
      interrupt_handler_create(SET_CURSOR_POS_INT, setcursor_handler) < 0 ||         /*Installing the set_cursor_pos interrupt */
      interrupt_handler_create(GET_CURSOR_POS_INT, getcursor_handler) < 0 ||         /*Installing the get_cursor_pos interrupt */
      interrupt_handler_create(PRINT_INT, print_handler) < 0)                        /*Installing the print interrupt */
  {
    return -1;                                                                       /*if installation of any of the interrups fail return -1*/
  }

  return 0;                                                                          /*return 0 otherwise */
}

//---------------------------------------------------------------------------------------------

void keyboard_handler(InterruptVector vector, Registers *registers){
  
  int ch;  
  
  ch = inb(KEYBOARD_PORT);       /* if keyboard interrupts...add the character to the buffer */
  
  disable_interrupts();          /* disable interrupts to make sure no one else interrups while char is being added */
  insert(ch);                    /* call the function insert to add char to buffer */
  enable_interrupts();

  resource_wake (&keyboardResource,1); /*wake the 1st thread that is waiting */
  
  
}

//---------------------------------------------------------------------------------------------
int process_scode(void){

  int read, s_code = -1;
   
  while(curr_size!=0 && s_code==-1){        /*Check to see that invalid ScanCodes are not returned*/
    disable_interrupts();                   /*Disable interrupsts before modifying buffer*/
    curr_size--;
    
    read=buffer[front];                     /*Get the character from the front of the queue*/

    buffer[front]=NULL;                     /*reset the front pointer*/
    front++;                                
    
    if(front==KEYBOARD_BUFFER_SIZE)         /* reset the front */
      front=0; 
    
    enable_interrupts();                   /*enable interrupts once the char has been pulled out of the buffer */
    
    s_code = process_scancode(read);        
  }
  
  return s_code;                            /*return the char read to the calling function*/
}

void getchar_wake_handler(Thread* thread, void *data){

  int s_code; 

  Registers * registers = (Registers *) data;
 
  s_code=process_scode();                            /*Get the next char from the buffer*/
  
  if(s_code!=-1)
    registers->EAX= s_code;                          /*return the char*/
  else                                               
  if(registers != &thread->registers)
    resource_wait(&keyboardResource, thread, getchar_wake_handler, &thread->registers);
  else
    resource_rewait(&keyboardResource, thread);      /*if no character; make the thread rewait */
}


void getchar_handler(InterruptVector vector, Registers *registers){
  
  Thread * thread = thread_current ();                   
  
  thread->registers = *registers;
  if (list_empty (& keyboardResource.waitingThreads))
    getchar_wake_handler(thread, registers);    /*call the wake_handler which takes care of i/o buffer manipulation */
  else
    resource_wait(&keyboardResource, thread, getchar_wake_handler, &thread->registers); /* go to resource wait of more tasks are already waiting */
}


void readline_handler(InterruptVector vector, Registers *registers){

  struct rl * args =(struct rl *) registers->ESI;           /*Get the parameters for readline */
  Thread * thread;

  if(VM_VALIDATE(args,struct rl,1)<0 ||                     /*Validate the memory address of the args passed in */
     args->len <= 0 || args->len > sizeof(rl_buffer) ||
     VM_VALIDATE_WRITE(args -> buf, char, args->len) < 0){    
    registers->EAX=-1;                                      /*If memory address are invalid return -1*/
    return; 
  }

  thread = thread_current ();
  thread -> registers = * registers;
  thread -> registers.EAX = args -> len;
  if (list_empty (& keyboardResource.waitingThreads))
  {
    thread->waitingHandler = readline_wake_handler;
    thread->waitingData = args -> buf;
    readline_wake_handler (thread, args -> buf);
    if (thread -> state == THREAD_STATE_RUNNING)
      registers -> EAX = thread -> registers.EAX;
  }
  else
    resource_wait(&keyboardResource, thread, readline_wake_handler, args -> buf);      /*Add the thread to the resource wait queue */
  return;
}

void readline_wake_handler(Thread* thread, void *data){
  Registers * registers = &thread->registers;
  int read,len = registers->EAX;
  char * buf = (char *)data;

  if(rl_thread != thread->id){                                /*check if the thread owns the readline buffer */    
    rl_index = 0;                                             /*reset the variable values */    
    rl_thread = thread->id;
  }	 
  do {      
    read=process_scode();                                     /*read the car from the buffer */
    if(read==-1)
      break;
    switch(read){
    case '\r':
      while(rl_index > 0){
	--rl_index;
	putbyte('\b');
      }
      break;
    case '\b':
      if (rl_index>0){
        --rl_index;
	putbyte('\b');
      }
      break;
    default:
      rl_buffer[rl_index]=read;                                 /*put the characters in the temp readline buffer */
      rl_index++;
      putbyte((char)read);
      break;
    }
  } while (rl_index < len && read != '\n');                  /*keep on readine char from buffer till either you get a newline char */
                                                             /*or the index exceeds the number of chars to be read */
          
  if(rl_index >= len || read=='\n'){                         /*if the one of the conditions is met */
    rl_thread = -1; /* release the readline buffer */

    vm_context_switch (thread -> task);                      /*context switch to the thread calling readline */

    if(VM_VALIDATE_WRITE(buf, char, len)<0){                    /*revalidate the args sent in */
      registers->EAX=-1;
    } else if(rl_index<=len){                                       /*check condition of how many characters to copy */
      memcpy(buf, rl_buffer, rl_index);                /*copy the string to the buffer */
      registers->EAX = rl_index;                             /*return the number of char's copied */
    } else {
      memcpy(buf, rl_buffer, len);
      registers->EAX = len;
    }

    vm_context_switch (task_current ());                     /*context switch to the current task */
    return;
  }    
  
  resource_rewait(&keyboardResource, thread);               /*if no newline char or buffer is not equal to len; rewait for interrupt*/
}
  
void print_handler(InterruptVector vector, Registers *registers){

  struct rl * args =(struct rl *) registers->ESI;          /*get the args */

  if(VM_VALIDATE(args,struct rl,1)<0 ||                    /*validate the args */
     args->len<0 ||
     VM_VALIDATE(args->buf, char, args->len)<0){
    registers->EAX=-1;                                     /*return if args are invalid */
    return;
  }

  putbytes(args->buf,args->len);                          /*call the console function to print to screen */

  registers->EAX=0;
}

void termcolor_handler(InterruptVector vector, Registers *registers){

  if(registers->ESI<0 || registers->ESI>0xFF){       /*validate the args */
    registers->EAX=-1;
    return;
  }

  set_term_color(registers->ESI);                   /*call the console function to set the term color */

  registers->EAX=0;
}

void setcursor_handler(InterruptVector vector, Registers *registers){
 
  struct sc *c = (struct sc *)registers->ESI;                /*get the args */
 
  if(VM_VALIDATE(c, struct sc, 1)<0 ||                       /*validate the args */
     c->row<0 || c->row>=CONSOLE_HEIGHT ||
     c->col<0 || c->col>=CONSOLE_WIDTH)

  {
    registers->EAX=-1;
    return;
  }

  set_cursor(c->row, c->col);                               /*call the console function to reset the cursor potion */

  registers->EAX=0;
}


void getcursor_handler(InterruptVector vector, Registers *registers){

  struct gc *c = (struct gc *)registers->ESI;                            /*get the args*/

  if(VM_VALIDATE(c, struct gc, 1)<0  || VM_VALIDATE_WRITE(c->row, int, 1)<0 ||  VM_VALIDATE_WRITE(c->col, int, 1)<0) { /*validate the args */
    registers->EAX=-1;
    return;
  }

  get_cursor((int*)c->row,(int*)c->col);                               /*call the console function to get the cursor potion */

  registers->EAX = 0;
}

//---------------------------------------------------------------------------------------------
void insert(int ch){
  
  if(curr_size==KEYBOARD_BUFFER_SIZE) /* if buffer is full...discard character and return */
    return; 

  back++;             /* otherwise add character to the buffer */
  if(back == KEYBOARD_BUFFER_SIZE)
    back=0;
  
  buffer[back]=ch;           /* add new char to the back of the buffer*/
  curr_size++;               /* update the buffer size */
  
}
