/** @file kernel.c
 *  @brief The kernel entry-point.
 *
 *  Initializes the various kernel modules and enters into the scheduler.
 * 
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Khan (hkhan)
 *
 */

/* -- Includes -- */

/* multiboot header file */
#include <multiboot.h>    /* for boot_info */

/* memory includes. */
#include <lmm.public.h>   /* for lmm_remove_free() */

/* x86 specific includes */
#include <x86/seg.h>      /* for install_user_segs() */
#include <x86/pic.h>      /* for pic_init() */
#include <x86/base_irq.h> /* for base_irq_master/slave */

#include <scheduler.h>
#include <frame.h>
#include <task.h>
#include <thread.h>
#include <interrupt.h>
#include <timer.h>
#include <syscall.h>
#include <io.h>

/*
 * state for kernel memory allocation.
 */
extern lmm_t malloc_lmm;

/*
 * Info about system gathered by the boot loader
 */
extern struct multiboot_info boot_info;


static void
kernel_panic (char * message)
{
	/* output an error message and just sit around, since an unmanageable error occurred */
	putbytes (message, strlen (message));
	putbyte ('\n');

	while (1);
}

/** @brief Kernel entrypoint.
 *  
 * Initializes the kernel modules and enters into the scheduler.
 * 
 *  @return Should not return
 */

int kernel_main()
{
    /*
     * Tell the kernel memory allocator which memory it can't use.
     * It already knows not to touch kernel image.
     */
    lmm_remove_free( &malloc_lmm, (void*)USER_MEM_START, USER_MEM_SIZE );
    lmm_remove_free( &malloc_lmm, (void*)0, 0x100000 );

    /*
     * initialize the PIC so that IRQs and
     * exception handlers don't overlap in the IDT.
     */
    pic_init( BASE_IRQ_MASTER_BASE, BASE_IRQ_SLAVE_BASE );

    /* initialize the console */
    show_cursor ();
    clear_console ();

    if (interrupt_dispatcher_initialize () < 0)
	kernel_panic ("Error initializing interrupt dispatcher");

    /* initialize the frame allocate with available high memory above the kernel heap */
    if (frame_allocator_initialize (USER_MEM_START / FRAME_SIZE, (boot_info.mem_upper + 1024) * 1024 / FRAME_SIZE - 1) < 0)
	kernel_panic ("Error initializing frame allocator");
    /* reserve the space between the low memory and high memory as unallocateable, just in case */ 
    frame_allocator_reserve (boot_info.mem_lower / FRAME_SIZE, 1024 * 1024 / FRAME_SIZE - 1);

    if (scheduler_initialize () < 0)
	kernel_panic ("Error initializing schedule");

    if (thread_manager_initialize () < 0)
	kernel_panic ("Error initializing thread manager");

    if (task_manager_initialize () < 0)
	kernel_panic ("Error initializing task manager");

    /* setup the syscall handlers */
    if (syscall_initialize () < 0)
	kernel_panic ("Error initializing system calls");

    /* initialize the io facilities */
    if (io_initialize () < 0)
	kernel_panic ("Error initializing input/output");

    /* disable interrupts so the timer does not get triggered prematurely */
    disable_interrupts ();

    /* start the timer services */
    if (timer_initialize () < 0)
	kernel_panic ("Error initializing timer");	 

    /* finally, enter into the init task */
    /* interrupts will be re-enabled when the initial context switch sets up EFLAGS */
    scheduler_jump ();

    kernel_panic ("Error jumping to init task");

    return 0;
}


