/** @file malloc_wrapper.c
 *  @brief Provides interrupt safe wrappers for memory allocation.
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */

#include <common.h>
#include <malloc.h>
#include <x86/proc_reg.h>

/* evaluates a given expression with interrupts disabled and returns a result of the given type */
#define WITH_INTERRUPTS_DISABLED_RETURN(type, expression) \
	type result; \
	disable_interrupts(); \
	result = expression; \
	enable_interrupts(); \
	return result;

/* evaluates a given expression with interrupts disabled */
#define WITH_INTERRUPTS_DISABLED(expression) \
        disable_interrupts(); \
        expression; \
        enable_interrupts();

/* safe versions of malloc functions that all do the obvious thing
 * rather, they disable interrupts and call the "_" versions of the functions they wrap */
void *malloc(size_t size)
{
	WITH_INTERRUPTS_DISABLED_RETURN (void *, _malloc (size))
}

void *memalign(size_t alignment, size_t size)
{
	WITH_INTERRUPTS_DISABLED_RETURN (void *, _memalign (alignment, size))
}

void *calloc(size_t nelt, size_t eltsize)
{
	WITH_INTERRUPTS_DISABLED_RETURN (void *, _calloc (nelt, eltsize))
}

void *realloc(void *buf, size_t new_size)
{
	WITH_INTERRUPTS_DISABLED_RETURN (void *, _realloc (buf, new_size))
}

void free(void *buf)
{
	WITH_INTERRUPTS_DISABLED (_free (buf))
}

void *smalloc(size_t size)
{
	WITH_INTERRUPTS_DISABLED_RETURN (void *, _smalloc (size))
}

void *smemalign(size_t alignment, size_t size)
{
	WITH_INTERRUPTS_DISABLED_RETURN (void *, _smemalign (alignment, size))
}

void *scalloc(size_t size)
{
	WITH_INTERRUPTS_DISABLED_RETURN (void *, _scalloc (size))
}

void sfree(void *buf, size_t size)
{
	WITH_INTERRUPTS_DISABLED (_sfree (buf, size))
}


