/** @file scheduler.c
 *  @brief Provides the implementantion of scheduler related functions
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */
#include <stdlib.h>
#include <scheduler.h>
#include <thread.h>
#include <interrupt.h>
#include <vm.h>
#include <x86/proc_reg.h>

static Scheduler scheduler;

int
scheduler_initialize (void)
{
  /*setup the scheduler */
  scheduler.currentThread = NULL;
  /*clear the list of ready threads */
  list_clear (& scheduler.readyThreads);
  
  return 0;
}

Thread *
thread_current (void)
{
  /*return the currently running thread */
  return scheduler.currentThread;
}

void
scheduler_tick (void)
{
  /*Update the ticks for the current running thread */
  if (scheduler.currentThread != NULL &&
      scheduler.currentThread -> quantum > 0)
    -- scheduler.currentThread -> quantum;
}

void
scheduler_context_switch (Registers * registers)
{
  disable_interrupts ();
  
  if (scheduler.currentThread != NULL)
    {
      /*return direclty if currentThread has more time to run*/
      if (scheduler.currentThread -> quantum > 0)
	{
	  enable_interrupts ();
	  return;
	}
      /*save the regististers and update the state of the current thread */
      scheduler.currentThread -> registers = * registers;
      scheduler.currentThread -> state = THREAD_STATE_READY;
      
      /*insert the current thread to the back of the list */
      list_insert_back (& scheduler.readyThreads, & scheduler.currentThread -> schedulerList);
    }
  else
    {
      while (list_empty (& scheduler.readyThreads))
	{
	  enable_interrupts ();
	  
	  interrupt_wait (); /*if list of threads is empty; busy wait */
			
	  disable_interrupts ();
	}
    }
  
  /*Makes the 1st thread in the ready list the currently running thread */
  scheduler.currentThread = LIST_ELEMENT (list_remove_front (& scheduler.readyThreads), Thread, schedulerList);
  
  /*context switch to the new therad */
  * registers = scheduler.currentThread -> registers;
  vm_context_switch (scheduler.currentThread -> task);
  
  /*chages teh state of the thread */
  scheduler.currentThread -> state = THREAD_STATE_RUNNING;
  /*update the quantam to run */
  scheduler.currentThread -> quantum = THREAD_QUANTUM_INITIAL;
  
  enable_interrupts ();
}

void
scheduler_jump (void)
{
	Registers registers;

	scheduler_context_switch (& registers); /*select the first thread to run */
	
	if (scheduler.currentThread != NULL)
	  registers_jump (& registers); /* jump into the thread */
}

void
scheduler_unschedule (Thread * thread)
{
	disable_interrupts ();

	if (scheduler.currentThread == thread)
	  scheduler.currentThread = NULL;            /*make the current thread NULL */
	else
	  list_remove (& thread -> schedulerList);   /*remove the thread from the ready list */

	thread -> state = THREAD_STATE_UNSCHEDULED;  /*chage the thread state to unscheduled */

	enable_interrupts ();
}

void
scheduler_schedule (Thread * thread)
{
	disable_interrupts ();

	if (thread -> quantum > 0)
	  list_insert_front (& scheduler.readyThreads, & thread -> schedulerList);  /*insert the thread to the front if it has more time to run */
	else
	  list_insert_back (& scheduler.readyThreads, & thread -> schedulerList);  /*insert at back if originally used up all its time */

	thread -> state = THREAD_STATE_READY; /*make the thread ready to run again */

	enable_interrupts ();
}

void
scheduler_yield (Thread * thread)
{
  disable_interrupts ();
  
  if (scheduler.currentThread != NULL)
    {
      list_insert_back (& scheduler.readyThreads, & scheduler.currentThread -> schedulerList);    /*insert the current thread to the back of the list */
      
      scheduler.currentThread -> state = THREAD_STATE_READY;   /*change the state of the thread */
      scheduler.currentThread = NULL;
    }
  
  if (thread != NULL && thread -> state == THREAD_STATE_READY)   /*If thread is ready to run then put it at the front of the ready list */
    {
      list_remove (& thread -> schedulerList);                                        
      list_insert_front (& scheduler.readyThreads, & thread -> schedulerList);
    }
  
  enable_interrupts ();
}

int
resource_create (Resource * resource)
{
  list_clear (& resource -> waitingThreads);     /*clear the waiting list*/

  return 0;
}

void
resource_destroy (Resource * resource)
{
}

void
resource_wait (Resource * resource, Thread * thread, void (* waitingHandler) (Thread *, void *), void * waitingData)
{
  scheduler_unschedule (thread);          /*unschedule the thread */
  
  disable_interrupts ();
  
  thread -> state = THREAD_STATE_WAITING;  /*change the state of the thread to waiting*/
  thread -> waitingHandler = waitingHandler;
  thread -> waitingData = waitingData;
  
  list_insert_back (& resource -> waitingThreads, & thread -> schedulerList); /*insert the thread at the back of the waiting list */
  
  enable_interrupts ();
}

void
resource_rewait (Resource * resource, Thread * thread)
{
  scheduler_unschedule (thread);   /*unschedule the thread */
  
  disable_interrupts ();
  
  thread -> state = THREAD_STATE_WAITING;  /*change the state of the thread to waiting*/
  
  list_insert_front (& resource -> waitingThreads, & thread -> schedulerList); /*insert the thread to the front of the waiting list */
  
  enable_interrupts ();
}

int
resource_wake (Resource * resource, int limit)
{
	ListIterator position;
	Thread * back;
	int awakened = 0;
	
	disable_interrupts ();

	if (list_empty (& resource -> waitingThreads))      /*check the list for the resource */
	{
		enable_interrupts ();               
	
		return awakened;            /*if no threads waiting then return */
	}

	back = LIST_ELEMENT (list_back (& resource -> waitingThreads), Thread, schedulerList);
	for (position = list_begin (& resource -> waitingThreads); limit < 0 || awakened < limit;) /*repeat this for the number of threads to be awakened*/
	{
		Thread * thread;

		position = list_next (position);
		thread = LIST_ELEMENT (list_remove (list_previous (position)), Thread, schedulerList);        /*Remove from the resourse list */

		enable_interrupts ();

		scheduler_schedule (thread);            /*schedule the thread */

		++ awakened;       /*update the counter of number of thread woken up*/

		if (thread -> waitingHandler != NULL)
		  (* thread -> waitingHandler) (thread, thread -> waitingData); /*calls the wake handler for the woken up thread */

		disable_interrupts ();

		if (thread == back)
		  break;
	}

	enable_interrupts ();

	/*return number of awakened threads */
	return awakened;
}	

 
