/** @file stubs.S
 * @brief Implements interrupt stubs. 
 * 
 * Implements the interrupt stubs that link to the interrupt dispatcher.
 * The stubs push the vector of the interrupt and a faux-error-code if
 * necessary, then finally move into the handlers which saves/restores user
 * level registers onto the interrupt stack. The registers and interrupt vector
 * are finally passed to the interrupt dispatcher. 
 *
 */

#define ASSEMBLER
#include <x86/seg.h>

/* generates the exception stub name for the given vector */
#define EXCEPTION_STUB_NAME(vector) exception_stub_ ## vector

/* generates the exception stub for the given vector 
 * the stub pushes a faux-error-code and the vector
 */
#define EXCEPTION_STUB(vector) \
	.global EXCEPTION_STUB_NAME(vector); \
EXCEPTION_STUB_NAME(vector):; \
	pushl	$0; \
	pushl	$vector; \
	jmp	exception_handler; \
	.align 16;

/* generates the exception stub for the given vector
 * the exception already pushes an error code, so just push the vector
 */
#define EXCEPTION_STUB_WITH_ERROR(vector) \
        .global EXCEPTION_STUB_NAME(vector); \
EXCEPTION_STUB_NAME(vector):; \
        pushl   $vector; \
        jmp     exception_handler; \
        .align 16;

/* generates the interrupt stub name for the given vector */
#define INTERRUPT_STUB_NAME(vector) interrupt_stub_ ## vector

/* generates the interrupt stub for the given vector
 * interrupt stubs only need push the vector since they can have no error code at all
 */
#define INTERRUPT_STUB(vector) \
	.global INTERRUPT_STUB_NAME(vector); \
INTERRUPT_STUB_NAME(vector):; \
	pushl	$vector; \
	jmp	interrupt_handler; \
	.align 16;

/* the array of all stubs from 0 to 100, where each stub is 16 bytes long */
	.global interrupt_stubs
interrupt_stubs:
EXCEPTION_STUB(0)
EXCEPTION_STUB(1)
EXCEPTION_STUB(2)
EXCEPTION_STUB(3)
EXCEPTION_STUB(4)
EXCEPTION_STUB(5)
EXCEPTION_STUB(6)
EXCEPTION_STUB(7)
EXCEPTION_STUB_WITH_ERROR(8)
EXCEPTION_STUB(9)
EXCEPTION_STUB_WITH_ERROR(10)
EXCEPTION_STUB_WITH_ERROR(11)
EXCEPTION_STUB_WITH_ERROR(12)
EXCEPTION_STUB_WITH_ERROR(13)
EXCEPTION_STUB_WITH_ERROR(14)
EXCEPTION_STUB(15)
EXCEPTION_STUB(16)
EXCEPTION_STUB_WITH_ERROR(17)
EXCEPTION_STUB(18)
EXCEPTION_STUB(19)
EXCEPTION_STUB(20)
EXCEPTION_STUB(21)
EXCEPTION_STUB(22)
EXCEPTION_STUB(23)
EXCEPTION_STUB(24)
EXCEPTION_STUB(25)
EXCEPTION_STUB(26)
EXCEPTION_STUB(27)
EXCEPTION_STUB(28)
EXCEPTION_STUB(29)
EXCEPTION_STUB(30)
EXCEPTION_STUB(31)
INTERRUPT_STUB(32)
INTERRUPT_STUB(33)
INTERRUPT_STUB(34)
INTERRUPT_STUB(35)
INTERRUPT_STUB(36)
INTERRUPT_STUB(37)
INTERRUPT_STUB(38)
INTERRUPT_STUB(39)
INTERRUPT_STUB(40)
INTERRUPT_STUB(41)
INTERRUPT_STUB(42)
INTERRUPT_STUB(43)
INTERRUPT_STUB(44)
INTERRUPT_STUB(45)
INTERRUPT_STUB(46)
INTERRUPT_STUB(47)
INTERRUPT_STUB(48)
INTERRUPT_STUB(49)
INTERRUPT_STUB(50)
INTERRUPT_STUB(51)
INTERRUPT_STUB(52)
INTERRUPT_STUB(53)
INTERRUPT_STUB(54)
INTERRUPT_STUB(55)
INTERRUPT_STUB(56)
INTERRUPT_STUB(57)
INTERRUPT_STUB(58)
INTERRUPT_STUB(59)
INTERRUPT_STUB(60)
INTERRUPT_STUB(61)
INTERRUPT_STUB(62)
INTERRUPT_STUB(63)
INTERRUPT_STUB(64)
INTERRUPT_STUB(65)
INTERRUPT_STUB(66)
INTERRUPT_STUB(67)
INTERRUPT_STUB(68)
INTERRUPT_STUB(69)
INTERRUPT_STUB(70)
INTERRUPT_STUB(71)
INTERRUPT_STUB(72)
INTERRUPT_STUB(73)
INTERRUPT_STUB(74)
INTERRUPT_STUB(75)
INTERRUPT_STUB(76)
INTERRUPT_STUB(77)
INTERRUPT_STUB(78)
INTERRUPT_STUB(79)
INTERRUPT_STUB(80)
INTERRUPT_STUB(81)
INTERRUPT_STUB(82)
INTERRUPT_STUB(83)
INTERRUPT_STUB(84)
INTERRUPT_STUB(85)
INTERRUPT_STUB(86)
INTERRUPT_STUB(87)
INTERRUPT_STUB(88)
INTERRUPT_STUB(89)
INTERRUPT_STUB(90)
INTERRUPT_STUB(91)
INTERRUPT_STUB(92)
INTERRUPT_STUB(93)
INTERRUPT_STUB(94)
INTERRUPT_STUB(95)
INTERRUPT_STUB(96)
INTERRUPT_STUB(97)
INTERRUPT_STUB(98)
INTERRUPT_STUB(99)

/* save all registers except ESI and EDI */
#define SAVE_REGISTERS \
	pushl	%ebp; \
	pushl	%ebx; \
	pushl	%edx; \
	pushl	%ecx; \
	pushl	%eax;

/* move the kernel stack segment selector into all the other data segment registers */
#define INIT_SEGS \
	movw	%ss, %ax; \
	movw	%ax, %ds; \
	movw	%ax, %es; \
	movw	%ax, %fs; \
	movw	%ax, %gs;

/* if transitioning back to user level, overwrite the segment registers with the user data segment selector */
#define RESTORE_SEGS \
	movl	32(%esp), %eax; \
	cmpl	$USER_CS_SEGSEL, %eax; \
	jne	0f; \
	movl	$USER_DS_SEGSEL, %eax; \
	movw	%ax, %ds; \
	movw	%ax, %es; \
	movw	%ax, %fs; \
	movw	%ax, %gs; \
0:

/* restore all the registers */
#define RESTORE_REGISTERS \
	popl	%eax; \
	popl	%ecx; \
	popl	%edx; \
	popl	%ebx; \
	popl	%ebp; \
	popl	%esi; \
	popl	%edi;

/* the common exception handler for the exception stubs */
	.global exception_handler
exception_handler:
/* error code and save EDI */
	xchgl	%edi, 4(%esp)
/* exception vector and save ESI */
	xchgl	%esi, 0(%esp)

	SAVE_REGISTERS
	INIT_SEGS
	
/* call exception_dispatch (vector, error, registers) */
	pushl	%esp
	pushl	%edi
	pushl	%esi
	call	exception_dispatch
	addl	$12, %esp

	RESTORE_SEGS
	RESTORE_REGISTERS

/* return from the exception */
	iret


/* the common interrupt handler for the interrupt stubs */
        .global interrupt_handler
interrupt_handler:
/* interrupt vector */
        xchgl   %edi, 0(%esp)
/* save esi */
	pushl	%esi

	SAVE_REGISTERS
	INIT_SEGS

/* call interrupt_dispatch (vector, registers) */
        pushl   %esp
        pushl   %edi
        call    interrupt_dispatch
        addl    $8, %esp

	RESTORE_SEGS
	RESTORE_REGISTERS

/* return from the interrupt */
        iret


/* jump into a set of registers by pretending to return from an interrupt */	
	.global registers_jump
registers_jump:
/* overwrite the current stack pointer with the register set's ESP */
	movl	4(%esp), %esp

	RESTORE_SEGS
	RESTORE_REGISTERS

/* finish loader the registers */
	iret

