/** @file syscall.c
 *  @brief Installs the syscalls 
 *
 *  This file provides the implementation for installing the syscalls 
 *  @author Lee Salzman (lsalzman)
 *  @bug No known bugs.
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <exec2obj.h>
#include <x86/proc_reg.h>
#include <kerndebug.h>
#include <syscall_int.h>
#include <syscall.h>
#include <interrupt.h>
#include <task.h>
#include <thread.h>
#include <timer.h>
#include <vm.h>


/** @brief The handler to get the id of the thread 
 *
 * get the thread id of the current task 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
gettid_handler (InterruptVector vector, Registers * registers)
{
  registers -> EAX = thread_current () -> id; /*puts the id of the current thread in EAX */
}



/** @brief The handler fork a child
 *
 * forks a new task 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
fork_handler (InterruptVector vector, Registers * registers)
{
	Task * task;

	thread_current () -> registers = * registers;

	task = task_fork ();        /*Fork a new task */

	if (task != NULL)
	  registers -> EAX = task -> mainThread -> id;  /*put the threadID of the main task in EAX*/
	else
	  registers -> EAX = -1;
}


/** @brief The handler for task_exec
 *
 * execs a new task 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
exec_handler (InterruptVector vector, Registers * registers)
{
  ExecArgs * args = (ExecArgs *) registers -> ESI;
  char ** execvec;	
  
  /*validate the input args */
  if (VM_VALIDATE (args, ExecArgs, 1) < 0 ||
      VM_VALIDATE_STRING (args -> execname, char) < 0 ||
      VM_VALIDATE_STRING (args -> execvec, char *) < 0)
    {
      registers -> EAX = -1;
      return;
    }
  
  for (execvec = args -> execvec; * execvec != NULL; ++ execvec)
    {
      if (VM_VALIDATE_STRING (* execvec, char) < 0)
	{
	  registers -> EAX = -1;
	  return;
	}
    }
  

  if (task_exec (task_current (), args -> execname, args -> execvec) >= 0)
    {
      /*copy the registers to the execed task */
      * registers = thread_current () -> registers;
    }
  else	
    {
      registers -> EAX = -1;
    }
}


/** @brief The handler for task_exit
 *
 * exits a thread
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
exit_handler (InterruptVector vector, Registers * registers)
{
  thread_exit (thread_current (), registers -> ESI); /*call thread_exit */
}


/** @brief The handler for wait
 *
 *  wakes the thread after a specified task has completed
 *
 *  @param thread	the thread woken up
 *  @param data		the argument to the wake handler
 *
 *  @return Void 
 */
static void
wait_wake_handler (Thread * thread, void * data)
{
  Registers * registers = (Registers *) data;
  int * status = (int *) registers -> ESI;
  Task * task = thread -> task;
  ListIterator position;
  
  
  for (position = list_begin (& task -> childTasks); position != list_end (& task -> childTasks); position = list_next (position))
    {
      Task * child = (Task *) position;
      
      if (list_empty (& child -> threads))
	{
	  vm_context_switch (task);       /*context switch to the waiting task */
	  
	  if (VM_VALIDATE_WRITE (status, int, 1) < 0)	 /*validate the satus */
	    {
	      registers -> EAX = -1;
	    }
	  else
	    {
	      registers -> EAX = child -> exitId;   /*put the thread id in EAX */
	      * status = child -> exitStatus;       /*put the exit status in the user memory */
	      
	      task_destroy (child);                 /*kill the task */
	    }
	  
	  vm_context_switch (task_current ());      /*context switch to the current task */
	  
	  return;
	}
    }
  
  resource_wait (& task -> exitResource, thread, wait_wake_handler, & thread -> registers);  /*wait on the resource */
}


/** @brief The handler for putting thread in wait state
 *
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
wait_handler (InterruptVector vector, Registers * registers)
{
	int * status = (int *) registers -> ESI;
	Thread * thread = thread_current ();
	Task * task = thread -> task;

	if (VM_VALIDATE_WRITE (status, int, 1) < 0 ||           /* validate the memory */
	    list_empty (& task -> childTasks))
	{
		registers -> EAX = -1;
		return;
	}

	thread -> registers = * registers;

	wait_wake_handler (thread, registers);           /*call the wait_wake_handler */
}


/** @brief The handler for forking a new therad
 *
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
thread_fork_handler (InterruptVector vector, Registers * registers)
{
  Thread * thread = thread_current (),
    * forked;
  
  thread -> registers = * registers;
  
  forked = thread_fork (thread -> task);         /*Fork a new thread */
  if (forked == NULL)
    {
      registers -> EAX = -1;                     /*if it fails...return -1*/
    }
  else
    {
      registers -> EAX = thread -> id;           /*otherwise return the id */
    }
}

/** @brief The handler for yield
 *
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
yield_handler (InterruptVector vector, Registers * registers)
{
	Thread * target = NULL,
	       * thread;

	if (registers -> ESI != -1)
	{
		disable_interrupts ();

		target = thread_find (registers -> ESI);
		if (target == NULL || target -> state != THREAD_STATE_READY)  /*check the status of the target thread */
		{
			enable_interrupts ();
			
			registers -> EAX = -1;       /*if invalid put -1 in EAX */
			return;
		}

		enable_interrupts ();
	}

	registers -> EAX = 0;

	thread = thread_current ();
	thread -> registers = * registers;

	scheduler_yield (target);              /*call the scheduler to yield to the target thread */
}



/** @brief The handler for descheule
 *
 *  deschedules the currently running thread 
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
deschedule_handler (InterruptVector vector, Registers * registers)
{
	Thread * thread = thread_current ();
	int * reject = (int *) registers -> ESI;

	if (VM_VALIDATE (reject, int, 1) < 0)     /*validate the args passed in */
	{
		registers -> EAX = -1;
		return;
	}

	registers -> EAX = 0;

	if (* reject != 0)
	  return;

	thread -> registers = * registers;

	scheduler_unschedule (thread);  /*tell the scheduler to deschedule the thread */
}


/** @brief The handler for make_runnable
 *
 *  changes the status of the thread to run state
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
make_runnable_handler (InterruptVector vector, Registers * registers)
{
	Thread * thread = thread_find (registers -> ESI);

	if (thread == NULL || thread -> state != THREAD_STATE_UNSCHEDULED)  /*check the current status of the thread */
	{
		registers -> EAX = -1;
		return;
	}

	scheduler_schedule (thread);    /*schedule the target thread */

	registers -> EAX = 0;
}

/** @brief The handler for get_ticks
 *
 *  Retruns the number of ticks
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
get_ticks_handler (InterruptVector vector, Registers * registers)
{
  registers -> EAX = get_ticks (); /*put the number of ticks in EAX */
}

/** @brief The handler for sleep
 *
 *  Puts the thread to sleep
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
sleep_handler (InterruptVector vector, Registers * registers)
{
	int ticks = registers -> ESI;
	Thread * thread;
	Timer * timer;

	registers -> EAX = 0;

	/*check the current status of ticks */
	if (ticks < 0)
	{
		registers -> EAX = -1;
		return;
	}
	else
	if (ticks == 0)
	  return;

	thread = thread_current ();

	timer = timer_create (ticks);   /*create a new timer */
	if (timer == NULL)
	{
		registers -> EAX = -1;
		return;
	}

	thread -> registers = * registers;

	resource_wait (& timer -> sleepResource, thread, NULL, NULL);  /*make the resource wait till woken up */

	timer_activate (timer);  /*start the timer */
}



/** @brief The handler for new_pages
 *
 *  Allocates new memory to the calling task
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
new_pages_handler (InterruptVector vector, Registers * registers)
{
	NewPagesArgs * args = (NewPagesArgs *) registers -> ESI;
	Task * task;
	MemoryRegion * region;

	/*validate the args passed in */
	if (VM_VALIDATE (args, NewPagesArgs, 1) < 0 ||
	    (((MemoryAddress) args -> addr) % FRAME_SIZE) != 0 ||
	    args -> len < 0 ||
	    (args -> len % FRAME_SIZE) != 0)
	{
		registers -> EAX = -1;
		return;
	}

	task = task_current ();
	/*create more space for the current task */
	region = memory_region_create (task,
			VM_PROTECTION_USER | VM_PROTECTION_READ | VM_PROTECTION_WRITE,
			(MemoryAddress) args -> addr,
			(MemoryAddress) args -> addr + args -> len - FRAME_SIZE,
			NULL);
	if (region == NULL)
	{ 
		registers -> EAX = -1;
		return;
	}
	/*map the memory to the task */
	if (vm_map (task, region) < 0)
	{
		memory_region_destroy (region);

		registers -> EAX = -1;
		return;
	}
	/*clears the new memory region */
	memset ((char *) region -> addressStart, 0, region -> frameCount * FRAME_SIZE);

	registers -> EAX = 0;
}


/** @brief The handler for remove_pages
 *
 *  Removes new memory to the calling task
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
remove_pages_handler (InterruptVector vector, Registers * registers)
{
	void * addr = (void *) registers -> ESI;
	Task * task = task_current ();
	MemoryRegion * region = memory_region_find (task, (MemoryAddress) addr);

	/*check for the region to be valid for removing pages */
	if (region == NULL || 
	    region -> faultHandler != NULL ||
	    (MemoryAddress) addr != region -> addressStart)
	{
		registers -> EAX = -1;
		return;
	}
	
	/*unmap memory*/
	vm_unmap (task, region);
	/*destroy the memory region */
	memory_region_destroy (region);

	registers -> EAX = 0;
}

/** @brief The handler for halt
 *
 *  shuts down the OS
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
halt_handler (InterruptVector vector, Registers * registers)
{
	SIM_halt ();
}


/** @brief The handler for ls
 *
 *  returns the names of the executable files
 *
 *  @param vector       The syscall interrupt
 *  @param registers    The syscall registers
 *
 *  @return Void 
 */
static void
ls_handler (InterruptVector vector, Registers * registers)
{
	LSArgs * args = (LSArgs *) registers -> ESI;
	int app, offset;

	/*validate the args */
	if (VM_VALIDATE (args, LSArgs, 1) < 0 || 
	    args -> size <= 0 ||
	    VM_VALIDATE_WRITE (args -> buf, char, args -> size) < 0)
	{
		registers -> EAX = -1;
		return;
	}

	offset = 0;
	/*goes through the task list and prints it out */
	for (app = 0; app <  exec2obj_userapp_count; ++ app)
	{
		const char * name = exec2obj_userapp_TOC[app].execname;

		if (offset + strlen (name) + 2 > args -> size)
		{
			registers -> EAX = -1;
			return;
		}

		strcpy (args -> buf + offset, name);

		offset += strlen (name) + 1;
	}

	args -> buf [offset] = '\0';
	
	registers -> EAX = app;
}
	


/** @brief Installs the system syscalls
 *  
 *  Sets up the interrupt handlers for the system calls.
 *
 *  @return 0 on success
 *         -1 on failure
 */
int
syscall_initialize (void)
{
  if (interrupt_handler_create (FORK_INT, fork_handler) == NULL ||                          /*intall the Fork handler */
      interrupt_handler_create (EXEC_INT, exec_handler) == NULL ||                          /*intall the Exec handler */
      interrupt_handler_create (GETTID_INT, gettid_handler) == NULL ||                      /*intall the Gettid handler */
      interrupt_handler_create (EXIT_INT, exit_handler) == NULL ||                          /*intall the Exit handler */
      interrupt_handler_create (WAIT_INT, wait_handler) == NULL ||                          /*intall the Wait handler */
      interrupt_handler_create (THREAD_FORK_INT, thread_fork_handler) == NULL ||            /*intall the Thread_fork handler */
      interrupt_handler_create (YIELD_INT, yield_handler) == NULL ||                        /*intall the Yield handler */
      interrupt_handler_create (DESCHEDULE_INT, deschedule_handler) == NULL ||              /*intall the Deschedule handler */
      interrupt_handler_create (MAKE_RUNNABLE_INT, make_runnable_handler) == NULL ||        /*intall the Make_Runnable handler */
      interrupt_handler_create (GET_TICKS_INT, get_ticks_handler) == NULL ||                /*intall the Get_Ticks handler */
      interrupt_handler_create (SLEEP_INT, sleep_handler) == NULL ||                        /*intall the Sleep handler */
      interrupt_handler_create (NEW_PAGES_INT, new_pages_handler) == NULL ||                /*intall the New_pages handler */
      interrupt_handler_create (REMOVE_PAGES_INT, remove_pages_handler) == NULL ||          /*intall the Remove_pages handler */
      interrupt_handler_create (HALT_INT, halt_handler) == NULL ||                          /*intall the Halt handler */
      interrupt_handler_create (LS_INT, ls_handler) == NULL)                                /*intall the Ls handler */
    {
      return -1;
    }
  
  return 0;
}


