/** @file task.c
 *  @brief Functions pretaining to the setup and execution of tasks
 *  
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */


#include <task.h>
#include <thread.h>
#include <vm.h>
#include <x86/proc_reg.h>
#include <malloc.h>
#include <elf.h>
#include <string.h>


static Task * initTask;

int
task_manager_initialize (void)
{
  char * argv [] = {"init", NULL};       
  
  initTask = task_create (NULL);          /*Creates a new task */
  
  if (vm_initialize (initTask) < 0)       /*Initializes the virtual memory for the task */
    return -1;
  
  if (task_exec (initTask, argv [0], argv) < 0)   /*Execs the task created */
    return -1;
  
  return 0;                                       /*return 0 on success */
}

Task *
task_current (void)
{
  Thread * thread = thread_current ();   /*Gets the currently running thread */
  
  if (thread != NULL)              
    return thread -> task;         /* returns the task of the currently running thread */
  
  return NULL;  /*return NULL if no task is running */
}

Task *
task_create (Task * parentTask)
{
  Task * task = (Task *) malloc (sizeof (Task));    /*Malloc size for the Task */
  
  if (task == NULL)                    /*If it fails return NULL */
	  return NULL;

  if (resource_create (& task -> exitResource) < 0)  /*If the call to create the resource fails retrun NULL */
    {
      free (task);
      
      return NULL;
    }

  /*Setup the task values in the struct */
  task -> parentTask = parentTask;
  list_clear (& task -> childTasks);
  task -> pageTable = NULL;
  list_clear (& task -> pageTables);
  list_clear (& task -> memoryRegions);
  task -> mainThread = NULL;
  list_clear (& task -> threads);
  task -> exitId = -1;
  task -> exitStatus = -1;


  
  if (parentTask != NULL)
    {
      disable_interrupts ();                               /*Disable interrupts to insert the task to the back of the list */
      list_insert_back (& parentTask -> childTasks, task); /*Insert the task */
      enable_interrupts ();                                /*Enable Interrupts*/
    }
  
  return task;                                             /*Return the task created */
}

void
task_destroy (Task * task)
{
  disable_interrupts ();  
  
  list_remove ((ListIterator) task);          /*Remove the task for the list of tasks */
  
  enable_interrupts ();
  
  free (task);                                /*Free the memory of the task */
}

void
task_exit (Task * task)
{
	disable_interrupts ();

	if (task != initTask)    /*move all the task's children to the init task if its not the init task */
	{
		while (! list_empty (& task -> childTasks))
		{
		  list_insert_back (& initTask -> childTasks, list_remove_front (& task -> childTasks));  /* Insert the task as children of the initTask*/
		}

		resource_destroy (& task -> exitResource);            /*Destroy the resource with the task*/
	}

	enable_interrupts ();

        while (! list_empty (& task -> threads))
        {
	  thread_destroy ((Thread *) list_front (& task -> threads));       /*Destroy the threads */
        }

	while (! list_empty (& task -> memoryRegions))
	{
	  memory_region_destroy ((MemoryRegion *) list_front (& task -> memoryRegions));  /*Destroy the memory regions */
	}	

	while (! list_empty (& task -> pageTables))
	{
	  frame_free ((Frame *) list_remove_front (& task -> pageTables));   /*Free the page tables */
	}

	task -> pageTable = NULL;
	task -> mainThread = NULL;

	if (task -> parentTask != NULL)
	  resource_wake (& task -> parentTask -> exitResource, 1); /*Wake the parent task */
	else
	  free (task);
}

Task *
task_fork (void)
{
  Task * parent = task_current (),   /*make the calling task the parent */
    * child;
  
  if (parent == NULL)    /*If parent doens't exist return NULL */
    return NULL;
  
  child = task_create (parent);  /*Create the new task */
  if (child == NULL)             /*Check if the child has been created properly */
    return NULL;
  
  if (vm_fork (child) < 0 ||
      thread_fork (child) == NULL)
    {
      task_exit (child);
      task_destroy (child);
      
      return NULL;
    }
  
  return child;                /*Return the neww task created */
}

static char execnameBuffer [EXECNAME_LIMIT];
static char execvecBuffer [EXECVEC_BUFFER_LIMIT];
static char * execvecArgv [EXECVEC_ARGV_LIMIT];
 
int
task_exec (Task * task, char * execname, char ** execvec)
{
  simple_elf_t elf;
  int argc, execvecArgvSize, execvecBufferSize;
  Thread * thread;
  MemoryRegion * text, * data, * rodata, * stack;
  MemoryAddress stackAddress;	
  
  if (list_size (& task -> threads) > 1)    /* refuse to exec if there are any other threads besides the current one */
    return -1;
  
  if (strlen (execname) + 1 > sizeof (execnameBuffer))  /*fail if there's not enough room to buffer the execname */
    return -1;
  
  strcpy (execnameBuffer, execname);                 /*copy the name of the task to the execBuffer */
  
  execvecBufferSize = 0;
  for (argc = 0; execvec [argc] != NULL; ++ argc)    /*copy the input params for the new task into temp buffers */
    {
      int argSize;
      
	/* fail if the number of args exceeds the buffering limit */
      if (argc + 1 >= EXECVEC_ARGV_LIMIT)
	return -1;
      
      argSize = strlen (execvec [argc]) + 1;
      
      execvecBufferSize += argSize;
      
	/* fail if there is not enough room to buffer the actual argument string */
      if (execvecBufferSize > sizeof (execvecBuffer))
	return -1;
      
      /* make the temporary argv buffer point to the locations the strings will actually be copied into in the task's new address space */
      execvecArgv [argc] = (char *) VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT - execvecBufferSize;
      
      /* copy the argument into the buffer */
      memcpy (execvecBuffer + sizeof (execvecBuffer) - execvecBufferSize, execvec [argc], argSize);
    }
  
  execvecArgv [argc] = NULL;
  execvecArgvSize = sizeof (char *) * (argc + 1);
  
  if (elf_load_helper (& elf, execnameBuffer) != ELF_SUCCESS)  /*check if the app exists */
    return -1;
  
  if (elf.e_txtstart > elf.e_rodatstart)   /*make sure the text region comes before the read-only data region */
    return -1;
  

  while (! list_empty (& task -> memoryRegions))  /*clearing up the old memory regions */
    {
      MemoryRegion * region = (MemoryRegion *) list_front (& task -> memoryRegions);
      
      /* unmap and destroy the regions */
      vm_unmap (task, region);
      
      memory_region_destroy (region);
    }
  
  
  /*create memory regions for the parts of the elf parts */
  if (((elf.e_txtstart + elf.e_txtlen) & ~ (FRAME_SIZE - 1)) == (elf.e_rodatstart & ~ (FRAME_SIZE - 1)))
    {
      /*if the text and rodata regions overlap on the same page */
      text = memory_region_create (task,
				   VM_PROTECTION_USER | VM_PROTECTION_READ | VM_PROTECTION_WRITE,
				   elf.e_txtstart & ~ (FRAME_SIZE - 1),
				   (elf.e_rodatstart + elf.e_rodatlen) & ~ (FRAME_SIZE - 1),
				   NULL);
      rodata = NULL;
    }
  else /*if the text & rodata section don't overlap on the same page  */
    { 
      text = memory_region_create (task,
				   VM_PROTECTION_USER | VM_PROTECTION_READ | VM_PROTECTION_WRITE,
				   elf.e_txtstart & ~ (FRAME_SIZE - 1), 
				   (elf.e_txtstart + elf.e_txtlen) & ~ (FRAME_SIZE - 1),
				   NULL);
      rodata = memory_region_create (task,
				     VM_PROTECTION_USER | VM_PROTECTION_READ | VM_PROTECTION_WRITE,
				     elf.e_rodatstart & ~ (FRAME_SIZE - 1),
				     (elf.e_rodatstart + elf.e_rodatlen) & ~ (FRAME_SIZE - 1),
				     NULL);
     
      /*if we fail to allocate a region then kill thread */
      if (rodata == NULL)
	{
	  if (thread_current () != NULL)
	    thread_exit (thread_current (), -1);
	  
	  return -1;
	}
    }
  

  /*create memory regions */
  data = memory_region_create (task,
			       VM_PROTECTION_USER | VM_PROTECTION_READ | VM_PROTECTION_WRITE,
			       elf.e_datstart & ~ (FRAME_SIZE - 1), 
			       (elf.e_datstart + elf.e_datlen + elf.e_bsslen) & ~ (FRAME_SIZE - 1),
			       NULL);
  stack = memory_region_create (task,
				VM_PROTECTION_USER | VM_PROTECTION_READ | VM_PROTECTION_WRITE,
				(VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT - execvecBufferSize - execvecArgvSize - FRAME_SIZE) & ~ (FRAME_SIZE - 1),
				VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT - FRAME_SIZE,
				stack_fault_handler);	
  stack -> growsDownward = VM_STACK_TOLERANCE;
  
  /*If we fail to allocate/map the regions, or read in the app data, then kill the thread*/
  if (text == NULL || data == NULL || stack == NULL ||
      vm_map (task, text) < 0 ||
      vm_map (task, data) < 0 ||
      (rodata != NULL && vm_map (task, rodata) < 0) ||
      vm_map (task, stack) < 0 ||
      getbytes (execnameBuffer, elf.e_txtoff, elf.e_txtlen, (char *) elf.e_txtstart) < 0 ||
      getbytes (execnameBuffer, elf.e_datoff, elf.e_datlen, (char *) elf.e_datstart) < 0 ||
      getbytes (execnameBuffer, elf.e_rodatoff, elf.e_rodatlen, (char *) elf.e_rodatstart) < 0)
    {
      if (thread_current () != NULL)
	thread_exit (thread_current (), -2);
      
      return -1;
    }
  
  
  /*zero out any data before the text section */
  memset ((char *) text -> addressStart, 0, elf.e_txtstart - text -> addressStart);
  if (rodata == NULL)
    {      
	/* the read-only data section and text section are in the same memory region, so zero out any gaps */
      memset ((char *) elf.e_txtstart + elf.e_txtlen, 0, elf.e_rodatstart - (elf.e_txtstart + elf.e_txtlen));
      memset ((char *) elf.e_rodatstart + elf.e_rodatlen, 0, text -> addressEnd + FRAME_SIZE - (elf.e_rodatstart + elf.e_rodatlen));
    }
  else
    {
	/* zero out any space before and after the text/rodata sections in their memory regions */
      memset ((char *) elf.e_txtstart + elf.e_txtlen, 0, text -> addressEnd + FRAME_SIZE - (elf.e_txtstart + elf.e_txtlen));
      memset ((char *) rodata -> addressStart, 0, elf.e_rodatstart - rodata -> addressStart);
      memset ((char *) elf.e_rodatstart + elf.e_rodatlen, 0, rodata -> addressEnd + FRAME_SIZE - (elf.e_rodatstart + elf.e_rodatlen));
      
    }
  /* zero out space before and after the data section in its memory region */ 
  memset ((char *) data -> addressStart, 0, elf.e_datstart - data -> addressStart);                  
  memset ((char *) elf.e_datstart + elf.e_datlen, 0, data -> addressEnd + FRAME_SIZE - (elf.e_datstart + elf.e_datlen));
  

  /* Disallow writing to the text section, and remap it with the new permissions */
  text -> protection = VM_PROTECTION_USER | VM_PROTECTION_READ;
  vm_map (task, text);
  
  /* if there is a separate rodata section, disallow writing and remap it with new permissions */
  if (rodata != NULL)
    {
      rodata -> protection = VM_PROTECTION_USER | VM_PROTECTION_READ;
      vm_map (task, rodata);
    }
  
  if (list_empty (& task -> threads))
    {
	/* if there is no thread in the task, then this is an init task, so create and schedule an initial thread */
      thread = thread_create (task);
      if (thread == NULL)
	return -1;
      
      scheduler_schedule (thread);
    }
  else
    {
	/* reuse the currently running thread */
      thread = thread_current ();
    }		
  
  /*zero out the memory on the stack */
  memset ((char *) stack -> addressStart, 0, stack -> frameCount * FRAME_SIZE);
  /*copy in the task args to the top of the stack */
  stackAddress = (stack -> addressEnd + FRAME_SIZE -  execvecBufferSize - execvecArgvSize) & ~ (sizeof (unsigned int) - 1);
  memcpy ((char **) stackAddress, execvecArgv, execvecArgvSize);	
  memcpy ((char *) VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT - execvecBufferSize, execvecBuffer + sizeof (execvecBuffer) - execvecBufferSize, execvecBufferSize);
  * ((unsigned int *) stackAddress - 1) = stackAddress;
  * ((unsigned int *) stackAddress - 2) = argc;
  * ((unsigned int *) stackAddress - 3) = 0;
  
  /* set the entry point and stack for the new thread */
  thread -> registers.EIP = elf.e_entry; 
  thread -> registers.ESP = stackAddress - sizeof (unsigned int) * 3;
  
  return 0;
}

