/** @file thread.c
 *  @brief Implementaion for threads
 *
 *  This file provides the implementation for threads
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */
#include <id.h>
#include <thread.h>
#include <register.h>
#include <scheduler.h>
#include <x86/proc_reg.h>
#include <x86/seg.h>
#include <malloc.h>
#include <string.h>

/*Thread ID Table */
static IdTable threadTable;

int
thread_manager_initialize (void)
{
  if (id_table_create (& threadTable, 1) < 0) /*create a thread id table */
    return -1; /*if it fails return -1*/
  
  return 0;
}

Thread *
thread_find (int id)
{
  return (Thread *) id_table_find (& threadTable, id); /*Find a thread in the IDTable */
}

Thread *
thread_create (Task * task)
{
  Thread * thread = (Thread *) malloc (sizeof (Thread));  /*Malloc size for a new thread */

  if (thread == NULL)
    return NULL;

  /*setup the values for the new thread */
  thread -> task = task;
  thread -> state = THREAD_STATE_UNSCHEDULED;
  thread -> quantum = 0;
  thread -> waitingHandler = NULL;
  thread -> waitingData = NULL;	
  
  memset (& thread -> registers, 0, sizeof (Registers));
  
  thread -> registers.CS = USER_CS_SEGSEL;
  thread -> registers.SS = USER_DS_SEGSEL;
  thread -> registers.EFLAGS = (1 << 1) /* reserved bit */
    | (1 << 9); /* interrupt flag */
  
  disable_interrupts ();

  /*insert the new thread into the id table */
  thread -> id = id_table_insert (& threadTable, thread);
  if (thread -> id < 0)
    {
      enable_interrupts ();
      
      free (thread);
      
      return NULL;
    }
  
  /*if thread list is empty */
  if (list_empty (& task -> threads))
    {
      /*update the task values */
      task -> exitId = thread -> id;
      task -> mainThread = thread;
    }
  
  /*insert at the back of the list */
  list_insert_back (& task -> threads, & thread -> threadList);
  
  enable_interrupts ();
  
  return thread;
}

void
thread_destroy (Thread * thread)
{
  /*check the state of the thread */
  scheduler_unschedule (thread);

  disable_interrupts ();
  
  /*remove thread from thread list */
  list_remove (& thread -> threadList);
  
  enable_interrupts ();
  
  /*free the memory for the thread */
  free (thread);
}

void
thread_exit (Thread * thread, int exitStatus)
{
  Task * task = thread -> task;
 
  /*Unschedule the thread */
  scheduler_unschedule (thread);
  
  /*Update the task */
  if (thread == task -> mainThread)
    {
      task -> mainThread = NULL;
    }
  
  /*change the staus of the task */
  task -> exitStatus = exitStatus;
  
  /*kill thre thread */
  thread_destroy (thread);
  
  /*if the list of threas is empty then exit the task */
  if (list_empty (& task -> threads))
    task_exit (task);
}

Thread *
thread_fork (Task * task)
{
	Thread * original = thread_current (),
		* clone;

	if (original == NULL)
	  return NULL;

	/*create a new thread in the task */
	clone = thread_create (task);
	if (clone == NULL)
	  return NULL;

	/*copy the values over to the new thread */
	clone -> registers = original -> registers;
	clone -> registers.EAX = 0;

	/*schedule the child */
	scheduler_schedule (clone);

	return clone;
}
	
