/** @file timer.c
 *  @brief Provides the implementantion of timer related functions
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */

#include <timer.h>
#include <interrupt.h>
#include <scheduler.h>
#include <timer_defines.h>
#include <x86/irq_list.h>
#include <x86/proc_reg.h>
#include <x86/pio.h>
#include <stdlib.h>
#include <malloc.h>

/*Number of ticks */
static unsigned int ticks = 0;
/*List of timers installed */
static List timers;

static void
timer_handler (InterruptVector vector, Registers * registers)
{
 
  /*update the number of ticks on each interrupt */
  ++ ticks;
  
  /*check to see if a timer expired */
  while (! list_empty (& timers))
    {
      Timer * timer = (Timer *) list_front (& timers);
      
      /*check to see if the first timer's deadline has passed*/
      if (ticks < timer -> deadline)
	break;
      
      enable_interrupts ();
      
	/* wake all threads waiting on the timer */
      resource_wake (& timer -> sleepResource, -1);
      
      disable_interrupts ();
      
      timer_destroy (timer);
    }
  
  scheduler_tick ();
}

int
timer_initialize (void)
{
	unsigned short timerPeriod;

	list_clear (& timers);

	/*Initialize the timer here */
	timerPeriod = TIMER_RATE / TIMER_HZ;
	outb (TIMER_MODE_IO_PORT, TIMER_SQUARE_WAVE);
	outb (TIMER_PERIOD_IO_PORT, timerPeriod & 0xFF);
	outb (TIMER_PERIOD_IO_PORT, timerPeriod >> 8);

	/*Install interrupt handler */
	if (interrupt_handler_create (INTERRUPT_VECTOR_INTERRUPTS + IRQ_TIMER, timer_handler) == NULL)
	  return -1;

	return 0;
}

unsigned int
get_ticks (void)
{
  /*return the number of ticks */
  return ticks;
}

Timer *
timer_create (unsigned int delta)
{
  
  /*create a new timer */
  Timer * timer = (Timer *) malloc (sizeof (Timer));

  
  if (timer == NULL)
    return NULL;
  
  /*The expiration of the timer */
  timer -> deadline = ticks + delta;
  
  /*Create a resource for the timer */
  resource_create (& timer -> sleepResource);
  
  return timer;
}

void
timer_activate (Timer * timer)
{
	ListIterator position;

	disable_interrupts ();

	/*TimerList setup in order of shortest time to deadline */
	for (position = list_begin (& timers);
	     position != list_end (& timers);
	     position = list_next (position))
	{
	  /*Getting the correct place */
		if (timer -> deadline < ((Timer *) position) -> deadline)
		  break;
	}

	/*Insert the timer in the list */
	list_insert (position, timer);

	enable_interrupts ();
}

void
timer_destroy (Timer * timer)
{
	disable_interrupts ();

	/*remove the timer from the list */
	list_remove ((ListIterator) timer);

	enable_interrupts ();
	/*destroy the resource */
	resource_destroy (& timer -> sleepResource);
	/*free the timer */
	free (timer);
}

