/** @file vm.c
 *  @brief Provides the implementantion of virtual memory management.
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Moin Khan (hkhan)
 *  @bug No known bugs.
 */

#include <x86/proc_reg.h>
#include <console.h>
#include <interrupt.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <frame.h>
#include <thread.h>
#include <task.h>
#include <vm.h>

static void
page_fault_handler (ExceptionVector vector, ExceptionData data, Registers * registers)
{
	static char buffer [512];
	Task * task = task_current ();
	/* the address of the page fault */
	MemoryAddress address = get_cr2 ();
	MemoryRegion * region;
	Thread * thread;

	/* look for a fault handler for the region, and if one is found and succeeds, do nothing */
	region = memory_region_find (task, address);
	if (region != NULL &&
	    region -> faultHandler != NULL &&
	    (* region -> faultHandler) (region, address, registers) >= 0)
	  return;

	/* if there is no fault handler, and a thread generated the page fault,
	 * then kill the thread and print a diagnostic message */
	thread = thread_current ();
	if (thread != NULL)
	{
		putbytes (buffer, snprintf (buffer, sizeof (buffer),
			"Thread %d killed by page fault\n"
			"  EIP: %x\n"
			"  CR2: %x\n",
			thread -> id, registers -> EIP, (unsigned) address));

		thread_exit (thread, -2);
	}
}	

int
vm_initialize (Task * task)
{
	MemoryAddress address, offset, end;
	PageTableEntry * pt0, * pt1;
	Frame * frame;

	/* install the page fault exception handler */
	if (exception_handler_create (14, page_fault_handler) == NULL)
	  return -1;

	/* allocate the init task's page directory */
	frame = frame_allocate (0);
	if (frame == NULL)
	  return -1;

	task -> pageTable = frame;

	list_insert_front (& task -> pageTables, frame);

	/* zero out the page directory */
	pt0 = (PageTableEntry *) FRAME_ADDRESS (frame);

        memset (pt0, 0, sizeof (PageTableEntry) * VM_PT0_SIZE);

	/* self-map the page directory at the top page directory entry */
        pt0 [VM_ADDRESS_PAGE_TABLE >> VM_PT0_SHIFT] = VM_PTE_PRESENT |
                                VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
                                FRAME_ADDRESS (frame);

	/* map in the linearly mapped kernel heap */
	for (address = VM_ADDRESS_KERNEL_HEAP;
	     address < VM_ADDRESS_KERNEL_HEAP + VM_ADDRESS_KERNEL_HEAP_LIMIT;
	     address += 1 << VM_PT0_SHIFT)
	{		
		/* allocate any page tables as necessary */
		frame = frame_allocate (0);
		if (frame == NULL)
		  return -1;

		pt0 [address >> VM_PT0_SHIFT] = VM_PTE_PRESENT |
				VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
				FRAME_ADDRESS (frame);

		/* linearly map the physical frames */
		pt1 = (PageTableEntry *) FRAME_ADDRESS (frame);
		for (offset = 0; offset < (1 << VM_PT0_SHIFT); offset += FRAME_SIZE)
		{
			pt1 [offset >> VM_PT1_SHIFT] = VM_PTE_PRESENT | VM_PTE_GLOBAL |
					VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
					(address + offset);
		}
	}
	
	/* map in the kernel stack address space */ 
       for (address = VM_ADDRESS_KERNEL_STACK + VM_ADDRESS_KERNEL_STACK_LIMIT - INTERRUPT_STACK_SIZE;
             address < VM_ADDRESS_KERNEL_STACK + VM_ADDRESS_KERNEL_STACK_LIMIT;
             )
        {
		/* allocate page tables */
                frame = frame_allocate (0);
                if (frame == NULL)
                  return -1;

                pt0 [address >> VM_PT0_SHIFT] = VM_PTE_PRESENT |
                                VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
                                FRAME_ADDRESS (frame);

		end = (address + (1 << VM_PT0_SHIFT)) & ~ ((1 << VM_PT0_SHIFT) - 1);
		if (end > VM_ADDRESS_KERNEL_STACK + VM_ADDRESS_KERNEL_STACK_LIMIT)
		  end = VM_ADDRESS_KERNEL_STACK + VM_ADDRESS_KERNEL_STACK_LIMIT;

		/* zero out the page table */
                pt1 = (PageTableEntry *) FRAME_ADDRESS (frame);
		memset (pt1, 0, sizeof (PageTableEntry) * ((address & VM_PT1_MASK) >> VM_PT1_SHIFT));
		memset (pt1 + ((end & VM_PT1_MASK) >> VM_PT1_SHIFT), 0, sizeof (PageTableEntry) * (VM_PT1_SIZE - ((end & VM_PT1_MASK) >> VM_PT1_SHIFT)));

		/* allocate frames for the stack */
		for (; address < end; address += FRAME_SIZE)
		{
			frame = frame_allocate (0);
			if (frame == NULL)
			  return -1;

                        pt1 [(address & VM_PT1_MASK) >> VM_PT1_SHIFT] = VM_PTE_PRESENT | VM_PTE_GLOBAL |
                                        VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
                                        FRAME_ADDRESS (frame);
                }
        }

	/* set the Page Global Enable flag */
	set_cr4 (get_cr4 () | (1 << 7));
	/* set the Page Directory Base Register to the initial level-0 page table */
	set_cr3 ((unsigned long int) pt0);
	/* set the enable-paging flag */
	set_cr0 (get_cr0 () | (1 << 31));

	return 0;
}

static unsigned int
unmapped_pt0_entries (PageTableEntry * pt0, MemoryAddress address, MemoryAddress end)
{
	unsigned int entries;

	/* search for any page directory entries within the memory region that are not present */

        for (entries = 0; address <= end; address += 1 << VM_PT0_SHIFT)
        {
                if (pt0 [address >> VM_PT0_SHIFT] & VM_PTE_PRESENT)
                  continue;
                
                ++ entries;
        }

	return entries;
}

static void
flush_tlb (void)
{
	/* modify the page directory base register, which implicitly flushes all non-global TLB entries */
	set_cr3 (get_cr3 ());
}

void
vm_context_switch (Task * task)
{
	/* set the page directory base register to the task's page directory */
	if (task != NULL)
	  set_cr3 (FRAME_ADDRESS (task -> pageTable));
}

static int
map_memory_region (Task * task, PageTableEntry * pt0, PageTableEntry * pt1, MemoryRegion * region, MemoryAddress start, MemoryAddress end)
{
	MemoryAddress address;
	ListIterator position;

	/* map the memory region into the task's page directory */
        for (address = start; address <= end; address += 1 << VM_PT0_SHIFT)
        {
		Frame * frame;

		if (pt0 [address >> VM_PT0_SHIFT] & VM_PTE_PRESENT)
		  continue;

		/* allocate a page table if none is present */
		frame = frame_allocate (0);
		if (frame == NULL)
		  return -1;

		/* insert the page table */
		list_insert_back (& task -> pageTables, frame);

		pt0 [address >> VM_PT0_SHIFT] = VM_PTE_PRESENT | 
						   VM_PROTECTION_USER |
						   VM_PROTECTION_READ | VM_PROTECTION_WRITE |
						   FRAME_ADDRESS (frame);

		/* clear the page table entries */
		memset (pt1 + (address >> VM_PT0_SHIFT) * VM_PT1_SIZE, 0, sizeof (PageTableEntry) * VM_PT1_SIZE);
        }
	
	/* map the actual frames beyond the starting address into the address space */
	address = region -> addressStart;
	for (position = list_begin (& region -> frames); position != list_end (& region -> frames); position = list_next (position))
	{
		Frame * frame = (Frame *) position;
 
		if (address > end)
		  break;

		if (address >= start)
		  pt1 [address >> VM_PT1_SHIFT] = VM_PTE_PRESENT | region -> protection | FRAME_ADDRESS (frame);

		address += FRAME_SIZE;
	}

	return 0;
}

static void
unmap_memory_region (PageTableEntry * pt0, PageTableEntry * pt1, MemoryRegion * region)
{
        MemoryAddress address;
 
	/* clear the present flag in all page existing page table entries within the memory region's bounds */
        for (address = region -> addressStart; address <= region -> addressEnd; address += 1 << VM_PT1_SHIFT)
        {
		if (pt0 [address >> VM_PT0_SHIFT] & VM_PTE_PRESENT)
                  pt1 [address >> VM_PT1_SHIFT] &= ~ VM_PTE_PRESENT;

        }
}

int
vm_map (Task * task, MemoryRegion * region)
{
	/* map the whole memory region into the current task's address space */
	if (map_memory_region (task, VM_PT0, VM_PT1 (0), region, region -> addressStart, region -> addressEnd) < 0)
	  return -1;

	/* flush the TLB to ensure the map takes effect with the right permissions */
	flush_tlb ();

	return 0;
}

int
vm_unmap (Task * task, MemoryRegion * region)
{
	/* unmap the whole memory region from the current task's address space */
	unmap_memory_region (VM_PT0, VM_PT1 (0), region);

	/* flush the TLB to remove any stale mappings */
	flush_tlb ();

	return 0;
}

int
vm_validate (MemoryAddress address, MemoryAddress end, unsigned int protection)
{
	/* fail if there are unmapped page tables within the range */
	if (unmapped_pt0_entries (VM_PT0, address, end) > 0)
	  return -1;

	/* ensure all PTEs within the memory range have a user protection level */
	while (address <= end)
	{
		if ((VM_PTE (address) & VM_PTE_PRESENT) == 0 ||
		    (VM_PTE (address) & (VM_PROTECTION_SUPERVISOR | VM_PROTECTION_USER)) != VM_PROTECTION_USER)
		  return -1;

		if ((VM_PTE (address) & protection) != protection)
		  return -1;

		address += FRAME_SIZE;
	}

	return 0;
}

int
vm_validate_string (MemoryAddress address, unsigned int type)
{
	int len = 0;

	/* only supports strings of type char, short, or int */
	if (type != 1 && type != 2 && type != 4)
	  return -1;

	/* search for the end of the string within user space */
	while (address >= VM_ADDRESS_USER_SPACE && 
		address < VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT)
	{
		MemoryAddress end = (address + FRAME_SIZE - 1) & ~ (FRAME_SIZE - 1);

		/* try and validate the current page we're searching on */
		if (vm_validate (address, address, VM_PROTECTION_READ) < 0)
		  return -1;

		/* check for the zero element within the current page */
		while (address + type <= end)
		{
			switch (type)
			{
			case 1:
				if (* (unsigned char *) address == 0)
				  return len;
				break;
			case 2:
				if (* (unsigned short *) address == 0)
				  return len;
				break;
			case 4:
				if (* (unsigned int *) address == 0)
				  return len;
				break;
			}
			++ len;
			address += type;
		}
			
		/* if the current element is not properly aligned, verify the next page is mapped */
		if (address < end)
		{
			if (vm_validate (end, end, VM_PROTECTION_READ) < 0)
			  return -1;

                        switch (type)
                        {
                        case 1:
                                if (* (unsigned char *) address == 0)
                                  return len;
                                break;
                        case 2:
                                if (* (unsigned short *) address == 0)
                                  return len;
                                break;
                        case 4:
                                if (* (unsigned int *) address == 0)
                                  return len;
                                break;
                        }
                        ++ len;
                        address += type;
		}
	}

	return -1;
}			
	
MemoryRegion *
memory_region_find (Task * task, MemoryAddress address)
{
        ListIterator position;

	/* search for the address within all the memory regions' bounds */
        for (position = list_begin (& task -> memoryRegions);
             position != list_end (& task -> memoryRegions);
             position = list_next (position))
        {
		MemoryRegion * region = (MemoryRegion *) position;

		/* modify the starting address by the growth tolerance */
                if (address >= region -> addressStart - region -> growsDownward && 
		    address < region -> addressEnd + FRAME_SIZE)
		  return region;
        }

	return NULL;
}

MemoryRegion *
memory_region_create (Task * task, unsigned int protection, MemoryAddress address, MemoryAddress end, MemoryFaultHandler faultHandler)
{
	ListIterator position;
	MemoryRegion * region;

	/* only allow memory region creation within user space */
	if (address < VM_ADDRESS_USER_SPACE || 
	    end >= VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT)
	  return NULL;

	/* if the memory region overlaps with existing regions, fail */
        for (position = list_begin (& task -> memoryRegions); 
             position != list_end (& task -> memoryRegions);
             position = list_next (position))
        {
                MemoryRegion * posRegion = (MemoryRegion *) position;

                if (end < posRegion -> addressStart)
		  break;

		if (address <= posRegion -> addressEnd ||
		    end <= posRegion -> addressEnd) 
		  return NULL;
	}
		  
	/* allocate and initialize the region */
	region = (MemoryRegion *) malloc (sizeof (MemoryRegion));
	if (region == NULL)
	  return NULL;

	region -> protection = protection;
	region -> growsDownward = 0;
	region -> addressStart = address;
	region -> addressEnd = end;
	region -> frameCount = (end - address) / FRAME_SIZE + 1;
	list_clear (& region -> frames);
	region -> faultHandler = faultHandler;

	/* allocate enough frames to fill the region, or fail if not available */
	while (address <= end)
	{
		Frame * frame = frame_allocate (0);
		if (frame == NULL)
		{
			while (! list_empty (& region -> frames))
			  frame_free ((Frame *) list_remove_front (& region -> frames));
	
			free (region);

			return NULL;
		}

		list_insert_back (& region -> frames, frame);

		address += FRAME_SIZE;
	}

	/* install the memory region into the task's region list */
	list_insert (position, region);

	return region;
}

void
memory_region_destroy (MemoryRegion * region)
{
	/* remove the region from its task's region list and free all the frames */
	list_remove ((ListIterator) region);

	while (! list_empty (& region -> frames))
	  frame_free ((Frame *) list_remove_front (& region -> frames));

	free (region);
}

static int 
fork_memory_regions (Task * dst, Task * src)
{
	ListIterator position;
	
	/* create memory regions with the same addresses and protections as one task in another */
	for (position = list_begin (& src -> memoryRegions);
	     position != list_end (& src -> memoryRegions);
	     position = list_next (position))
	{
		MemoryRegion * srcRegion = (MemoryRegion *) position,
		             * dstRegion = memory_region_create (dst, srcRegion -> protection, srcRegion -> addressStart, srcRegion -> addressEnd, srcRegion -> faultHandler);

		if (dstRegion == NULL)
		  return -1;
	}

	return 0;
}

static void
initialize_pt0 (PageTableEntry * pt0, Frame * frame)
{
	/* zero out the page directory */
        memset (pt0, 0, sizeof (PageTableEntry) * VM_PT0_SIZE);

	/* copy in the kernel heap mapping to the page directory */
        memcpy (pt0 + (VM_ADDRESS_KERNEL_HEAP >> VM_PT0_SHIFT),
                VM_PT0 + (VM_ADDRESS_KERNEL_HEAP >> VM_PT0_SHIFT),
                VM_ADDRESS_KERNEL_HEAP_LIMIT >> VM_PT0_SHIFT);

	/* self-map the page directory to itself */
	pt0 [VM_ADDRESS_PAGE_TABLE >> VM_PT0_SHIFT] = VM_PTE_PRESENT |
			VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
			FRAME_ADDRESS (frame);

	/* copy the stack address space mapping */
        pt0 [VM_ADDRESS_KERNEL_STACK >> VM_PT0_SHIFT] =
                VM_PT0 [VM_ADDRESS_KERNEL_STACK >> VM_PT0_SHIFT];
}

static void
copy_user_space (PageTableEntry * pt0)
{
	MemoryAddress address;

	/* copy all the data in the current address space to the target page table, utilizing the first
         * temporary mapping space
         */
	for (address = VM_ADDRESS_USER_SPACE;
	     address < VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT;
	     address += 1 << VM_PT0_SHIFT)
	{
		MemoryAddress offset;

		/* skip any unmapped superpages */
		if ((VM_PT0 [address >> VM_PT0_SHIFT] & VM_PTE_PRESENT) == 0)
		  continue;
		
		/* otherwise, put the superpage into the temporary mapping space */
		VM_PT0 [VM_ADDRESS_TEMPORARY_MAPPING_1 >> VM_PT0_SHIFT] = 
			VM_PTE_PRESENT |
			VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
			(pt0 [address >> VM_PT0_SHIFT] & (VM_PT1_MASK | VM_PT0_MASK));

		/* make sure the temporary mapping space is properly entered into the TLB */
		flush_tlb ();

		/* copy data from the current superpage into the temporary mapping space */
		for (offset = 0; offset < (1 << VM_PT0_SHIFT); offset += FRAME_SIZE)	
		{
			if ((VM_PTE (address + offset) & VM_PTE_PRESENT) == 0)
			  continue;

			memcpy ((void *) (VM_ADDRESS_TEMPORARY_MAPPING_1 + offset),
				(void *) (address + offset),
				FRAME_SIZE);
		}
	}
}
	
int
vm_fork (Task * task)
{
	ListIterator position;
	Frame * frame;

	/* fork all the memory regions from the current task into the new task */
	if (fork_memory_regions (task, task_current ()) < 0)
	  return -1;

	/* allocate a new page directory and install it */
	frame = frame_allocate (0);
	if (frame == NULL)
	  return -1;

	task -> pageTable = frame;

	list_insert_front (& task -> pageTables, frame);

	/* set the first temporary page directory mapping to the page directory */
	VM_PTE (VM_ADDRESS_TEMPORARY_MAPPING_1_PT0) =
		VM_PTE_PRESENT | 
		VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE | 
		FRAME_ADDRESS (frame);

	/* map the entire new task's page table structure into the second temporary mapping space */
        VM_PT0 [VM_ADDRESS_TEMPORARY_MAPPING_1 >> VM_PT0_SHIFT] =
                VM_PTE (VM_ADDRESS_TEMPORARY_MAPPING_1_PT0);

	/* ensure the mapping takes effect in the PTE */
	flush_tlb ();

	/* initialize the global mappings for the new page directory */
	initialize_pt0 ((PageTableEntry *) VM_ADDRESS_TEMPORARY_MAPPING_1_PT0, frame);

	/* map the new task's memory regions into its address space */
	for (position = list_begin (& task -> memoryRegions);
	     position != list_end (& task -> memoryRegions);
	     position = list_next (position))
	{
		MemoryRegion * region = (MemoryRegion *) position;

		if (map_memory_region (task, (PageTableEntry *) VM_ADDRESS_TEMPORARY_MAPPING_1_PT0, (PageTableEntry *) VM_ADDRESS_TEMPORARY_MAPPING_1, region, region -> addressStart, region -> addressEnd) < 0)
		  return -1;
	}

	/* copy all data from the current address space into the new task's regions */
	copy_user_space ((PageTableEntry *) VM_ADDRESS_TEMPORARY_MAPPING_1_PT0);

	return 0;
}

int
vm_copy (Task * task, MemoryAddress dst, char * src, unsigned int len)
{
	PageTableEntry * pt0;
	unsigned int copied;

	/* Since this may need to be called from anywhere, ensure only one thread is in this
	 * function at a time.
	 */
	disable_interrupts ();

	/* map the tasks page directory into the second temporary page directory mapping */
        VM_PTE (VM_ADDRESS_TEMPORARY_MAPPING_2_PT0) =
                VM_PTE_PRESENT |
                VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
		FRAME_ADDRESS (task -> pageTable);

	/* ensure the mapping takes effect */ 
	flush_tlb ();

	pt0 = (PageTableEntry *) VM_ADDRESS_TEMPORARY_MAPPING_2_PT0;

	/* go through all the superpages encompassing the source data
	 * and copy them into the target address space by mapping each superpage
	 * of the target address space at the second temporary mapping space
	 */ 
	while (len > 0)
	{
		/* map in the destination superpage to the second temporary mapping space */
                VM_PT0 [VM_ADDRESS_TEMPORARY_MAPPING_2 >> VM_PT0_SHIFT] =
                        VM_PTE_PRESENT |
                        VM_PROTECTION_SUPERVISOR | VM_PROTECTION_READ | VM_PROTECTION_WRITE |
                        (pt0 [dst >> VM_PT0_SHIFT] & (VM_PT1_MASK | VM_PT0_MASK));

		/* ensure the temporary mapping space properly maps into the TLB */
                flush_tlb ();

		/* now copy the superpages */
		copied = (1 << VM_PT0_SHIFT) - (dst & ~ VM_PT0_MASK);
		if (copied > len)
		  copied = len;

		memcpy ((char *) VM_ADDRESS_TEMPORARY_MAPPING_2 + (dst & ~ VM_PT0_MASK),
			src, copied);

		len -= copied;
		dst += copied;
		src += copied;
	}

	enable_interrupts ();

	return 0;		
}

int
stack_fault_handler (MemoryRegion * region, MemoryAddress address, Registers * registers)
{
	Task * task = task_current ();
	MemoryRegion * lowerBound;
	MemoryAddress oldStart;

	/* if the fault was outside user space, fail */
	if (address < VM_ADDRESS_USER_SPACE ||
	    address >= VM_ADDRESS_USER_SPACE + VM_ADDRESS_USER_SPACE_LIMIT)
	  return -1;

	/* if the fault extends into a previous memory region, fail */
	lowerBound = (MemoryRegion *) list_previous ((ListIterator) region);
	if (address < lowerBound -> addressEnd + FRAME_SIZE)
	  return -1;
	
	/* grow the start of the memory region to encompass the fault address */
	oldStart = region -> addressStart - FRAME_SIZE;
 
	while (region -> addressStart > address)
	{
		Frame * frame = frame_allocate (0);
		if (frame == NULL)
		  return -1;

		list_insert_front (& region -> frames, frame);

		region -> addressStart -= FRAME_SIZE;
	}

	/* map in the new frames */
	if (map_memory_region (task, VM_PT0, VM_PT1 (0), region, region -> addressStart, oldStart) < 0)
          return -1;

	/* zero out the new stack space */
	memset ((char *) region -> addressStart, 0, oldStart - region -> addressStart + FRAME_SIZE);

	return 0;
}
		
