/** @file spinlock.h
 *  @brief Interface for spinlocks.
 *
 *  This file provides simple spinlocks that repeatedly attempt to lock
 *  until they succeed. Spinlocks are simply integers holding a non-zero
 *  value if they are locked, or 0 if they are unlocked. Attempting to lock
 *  a spinlock busy-waits until locking succeeds. These are used to protect
 *  small and quickly exited sections of code in multiprocessor and interrupt
 *  safe manner.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __SPINLOCK_H__
#define __SPINLOCK_H__

typedef int spinlock_t;

#ifdef __GNUC__
#define INLINE		static __inline__
#define ASSEMBLY    	__asm__ __volatile__
#else /* ! __GNUC__ */
#error Spinlocks are only supported in GNU C.
#endif /*  __GNUC__ */

/** @brief Attempt to lock a spinlock.
 *
 *  Atomically tries to lock the spinlock.
 *
 *  @param lock a spinlock
 *
 *  @return the previous value of the lock,
 *          1 if locking failed (already locked)
 *          or 0 if locking suceeded (was unlocked)
 */
INLINE int 
spinlock_try_lock (spinlock_t * lock)
{
   int result = 1;

   /* use the atomic x86 xchg instruction to set the lock's value */
   ASSEMBLY ("xchgl %1, %0"
         : "=r" (result), "=m" (* lock));

   return result;
}

/** @brief Repeatedly attempt to lock spinlock until success.
 *
 *  Repeatedly try to lock the spinlock, effectively busy-waiting,
 *  until locking finally succeeds.
 *
 *  @param lock a spinlock
 *
 *  @return Void
 */
INLINE void
spinlock_lock (spinlock_t * lock)
{
   while (spinlock_try_lock (lock) != 0);
}

/** @brief Reset the spinlock.
 *
 *  Sets the spinlock to 0.
 *
 *  @param lock a spinlock
 *
 *  @return Void
 */
INLINE void
spinlock_unlock (spinlock_t * lock)
{
   * lock = 0;
}

#endif /* __SPINLOCK_H__ */
 
