/** @file threadlib.h
 *  @brief Thread library implementation details.
 *
 *  This file contains any definitions needed to implement thread
 *  library internals.
 *
 *  @author Lee Salzman (lsalzman)
 */

#ifndef __THREADLIB_H__
#define __THREADLIB_H__

#include <spinlock.h>

struct thread_context;

/** @brief A thread resource.
 *
 *  Implements a lockable resource which threads
 *  may block/wait upon and be subsequently awakened
 *  using a queue.
 */
typedef struct
{
	/* the first thread waiting in the queue, or NULL if empty */
	struct thread_context * waiting;
	/* the last thread waiting in the queue, or NULL if empty */
	struct thread_context * lastWaiting;
	/* the lock guarding access to the resource */
	spinlock_t lock;
} thread_resource_t;

/** @brief A user-land thread context.
 *
 *  A context storing relevant user-land thread
 *  meta-data.
 */ 
typedef struct thread_context
{
	/* synchronization resource for any specific join requests */
	thread_resource_t resource;
	/* next thread waiting in a resource queue if this thread is blocked */
	struct thread_context * nextWaiting;
	/* the user-land id of the thread */
	int userId;
	/* the kernel id of the thread */
	int systemId;
	/* the id of a thread that has joined with this thread */
	int joinId;
	/* whether or not the thread should be running (not descheduled) */
	int running;
	/* either the stack allocated for the thread if it is a running forked thread,
	 * THREAD_STACK_REUSED if it is the running master thread,
	 * or NULL if the thread has exited and is waiting for a join request.
	 */
	void * stack;
	/* the exit status of the thread if it has exited */
	void * exitStatus;
} thread_context_t;

/** @brief The base of any forked thread's stack.
 *
 *  Defines some meta-data and any arguments passed
 *  to the thread's entry function.
 */
typedef struct
{
	/* the forked thread */
	thread_context_t * thread;
	/* the entry function for the thread, passed as argument to thr_entry */
	void * (* entryFunction) (void *);
	/* the entry argument for the thread, passed as argument to thr_entry */
	void * entryArgument;	
} thread_stack_t;

#define THREAD_STACK_REUSED ((void *) ~0)

extern void thread_resource_init (thread_resource_t *);
extern void thread_resource_sleep (thread_resource_t *, thread_context_t *);
extern void thread_resource_wake (thread_resource_t *, int);

extern thread_context_t * thread_context_current (void);

extern void malloc_init (void);

/** @brief Forks a new thread.
 *
 *  Forks a new thread. The stack pointer is set to be to
 *  the new top of stack, and the code entry point is subsequently
 *  called. The exact paramaters and return value of the thread's
 *  entry point function are at the user's discretion, as the stack
 *  should be manipulated appropriately to pass parameters to the called
 *  entry point.
 *
 *  @param code the code entry point for the new thread
 *  @param stack the top of the new stack for the new thread
 *
 *  @return the id of the new thread, or -1 on error
 */
extern int thread_fork (void * code, void * stack);

#endif /* __THREADLIB_H__ */
