/** @file mutex.c
 *  @brief Implementation of mutexes.
 *  
 *  Implements mutexes as a thread resource
 *  with an additional locking depth (in case mutexes
 *  one day need to be recursive).
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Khan (hkhan)
 */ 

#include <stdlib.h>
#include <mutex.h>

int 
mutex_init (mutex_t * mp)
{
	/* initialize the mutex's resource and reset the lock depth */
	thread_resource_init (& mp -> resource);
  
	mp -> lockDepth = 0;

	return 0;
}


int
mutex_destroy (mutex_t * mp)
{
	/* acquire the resource */
	spinlock_lock (& mp -> resource.lock);

	/* if the mutex is locked or there are threads
	 * waiting to lock it, error
	 */
	if (mp -> lockDepth > 0 ||
	    mp -> resource.waiting != NULL)
	{
		spinlock_unlock (& mp -> resource.lock);

		return -1;
	}

	spinlock_unlock (& mp -> resource.lock);

	return 0;
}


int 
mutex_lock (mutex_t * mp)
{
	/* acquire the resource */
	spinlock_lock (& mp -> resource.lock);

	/* if the mutex is locked (non-zero lock depth),
	 * then sleep on the resource until it is not
	 */
	if (mp -> lockDepth > 0)
	  thread_resource_sleep (& mp -> resource, thread_context_current ());

	/* increment the lock depth */
	++ mp -> lockDepth;

	spinlock_unlock (& mp -> resource.lock);

	return 0;
}

int 
mutex_unlock (mutex_t * mp)
{
	/* acquire the resource */
	spinlock_lock (& mp -> resource.lock);

	/* if the mutex is not locked, error */
	if (mp -> lockDepth <= 0)
	{
		spinlock_unlock (& mp -> resource.lock);

		return -1;
	}

	/* decrement the lock depth */
	-- mp -> lockDepth;

	/* if the mutex is unlocked, wake up a thread sleeping
	 * on the resource, if any 
	 */
	if (mp -> lockDepth <= 0)
	  thread_resource_wake (& mp -> resource, 1);

	spinlock_unlock (& mp -> resource.lock);

	return 0;
}

