/** @file resource.c
 *  @brief Generalized synchronization resources for threads.
 *
 *  Provides generalized thread resources which encompass any
 *  resource a thread may need to block upon. Resources provide
 *  spinlocks to guard access to the resource as well as a
 *  blocking/waiting queue so that threads can be descheduled
 *  or made runnable again as necessary.
 *
 *  @author Lee Salzman (lsalzman)
 */
#include <stdlib.h>
#include <syscall.h>
#include <threadlib.h>

/** @brief Initialize a thread resource.
 *
 *  Initialize the thread resource's lock and wait queue.
 *
 *  @param resource the thread resource to be initialized
 *
 *  @return Void
 */
void
thread_resource_init (thread_resource_t * resource)
{
	/* clear the wait queue */
	resource -> waiting = NULL;
	resource -> lastWaiting = NULL;

	/* initialize the resource lock */
	resource -> lock = 0;
}

/** @brief Put a thread to sleep on a resource.
 *
 *  Adds the thread's context to the resource's wait
 *  queue and deschedules it. The resource must be locked
 *  before calling this function to avoid race conditions.
 *  The lock is released and the thread descheduled in an
 *  "atomic" fashion, much like a condition, and reacquired
 *  upon waking.
 *
 *  @param resource the resource to sleep on
 *  @param thread the thread context to be descheduled
 *
 *  @return Void
 */
void
thread_resource_sleep (thread_resource_t * resource, thread_context_t * thread)
{
	/* mark that the thread is about to be descheduled */
	thread -> running = 0;

	/* add the context to the wait queue */
	if (resource -> waiting == NULL)
	{
		/* put the thread at the head of the queue, as it is empty */
		resource -> waiting = thread;
	}
	else
	{
		/* put the thread after the current last waiting thread */
		resource -> lastWaiting -> nextWaiting = thread;
	}

	/* position the thread as the last in the queue */
	thread -> nextWaiting = NULL;
	resource -> lastWaiting = thread;

	/* release the resource */
	spinlock_unlock (& resource -> lock);

	/* deschedule the thread only if it is not supposed to be running */
	deschedule (& thread -> running);

	/* reacquire the resource */
	spinlock_lock (& resource -> lock);
}

/** @brief Wake threads waiting on a resource.
 *
 *  Wakes a specified number of threads and removes them from the resource's
 *  wait queue. The resource must be locked before calling this function so that
 *  threads are made runnable "atomically" with respect to thread_resource_sleep.
 *
 *  @param resource the resource to wake threads on
 *  @param count wake at most count number of threads if count >= 0, or otherwise wake all threads
 *
 *  @return Void
 */
void
thread_resource_wake (thread_resource_t * resource, int count)
{
	thread_context_t * thread;

	/* wake up threads until the count is exhausted or no threads
         * are waiting
         */
	while (count != 0 && resource -> waiting != NULL)
	{
		/* dequeue the first thread */
		thread = resource -> waiting;
		resource -> waiting = thread -> nextWaiting;

		/* clear the queue if it is empty */
		if (resource -> waiting == NULL)
		  resource -> lastWaiting = NULL;

		/* mark that the thread is about to be made runnable */
		thread -> running = 1;

		/* make the thread runnable */
		make_runnable (thread -> systemId);

		/* decrement the count if a limit was specified */
		if (count > 0)
		  -- count;
	}
}

