/** @file rwlock.c
 *  @brief Implementation of reader/write locks.
 *  
 *  Implements reader/writer locks using mutexes 
 *  and condition variables
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Khan (hkhan)
 */ 
#include <stdlib.h>
#include <rwlock.h>


int rwlock_init( rwlock_t *rwlock ){
  
  if(rwlock ==NULL)
    return -1;
  
  /*Initialize the condition vars and the mutexes*/
  rwlock->rdcnt=0;
  rwlock->wcnt=0;
  rwlock->writing=0;
  cond_init(&rwlock->conw);
  cond_init(&rwlock->conr);
  mutex_init(&rwlock->rwmut);
  
  return 0;
}


int rwlock_lock( rwlock_t *rwlock, int type ){
  
  if(rwlock==NULL)
    return -1;
  
  /*lock the resource*/
  mutex_lock(&rwlock->rwmut);
  
  if(type == RWLOCK_WRITE){
    /*Add a new writer*/
    rwlock->wcnt++;
    
    /*If a thread is already writing to resource then wait */
    while(rwlock->writing ==1 || rwlock->rdcnt>0)
      cond_wait(&rwlock->conw, &rwlock->rwmut);
 
    /*Grant the writer access*/
    rwlock->wcnt--;
    rwlock->writing = 1;
  }
  else if(type == RWLOCK_READ){
    /*Wait until there are no waiting writers to avoid deadlocks */
    while(rwlock->writing ==1 || rwlock->wcnt>0)
      cond_wait(&rwlock->conr, &rwlock->rwmut);
  
    /*Add a new writer*/
    rwlock->rdcnt++; 
  }
    
  /*Unlock Mutex*/
  mutex_unlock(&rwlock->rwmut);	 
  
  return 0;
}


int rwlock_unlock( rwlock_t *rwlock ){
  
  if(rwlock == NULL) 
    return -1; 
   
  /*Check if the lock is already in use*/
  if(rwlock->writing ==0 && rwlock->rdcnt==0)
    return -1;

  mutex_lock(&rwlock->rwmut); 
  

  if(rwlock->writing ==1){
    /* preferentially signal writers to avoid deadlocks*/
    if(rwlock->wcnt>0)
      cond_signal(&rwlock->conw);
    else
      cond_broadcast(&rwlock->conr);
    rwlock->writing =0;
  }
  else{
    rwlock->rdcnt--;
    if(rwlock->rdcnt==0)
      cond_signal(&rwlock->conw);
  }
    
  mutex_unlock(&rwlock->rwmut); 
  return 0;   
}


int rwlock_destroy( rwlock_t *rwlock ){
  
  /*check if resource is in use*/
  if(rwlock == NULL ||
     rwlock->writing==1 ||
     rwlock->rdcnt>0 ||
     mutex_destroy(&rwlock->rwmut)<0 ||
     cond_destroy(&rwlock->conw)<0 ||
     cond_destroy(&rwlock->conr)<0) 
    return -1;

  
  return 0;   
}
