/** @file sem.c
 *  @brief Implements semaphores.
 *
 *  Implements semaphores using mutexes 
 *  and condition variables. A reference
 *  count is used to count how many times
 *  the semaphore is signaled. The mutex
 *  guards access to the semaphore, while
 *  the condition variable is used to handle
 *  waiting for the reference count to increment.
 *
 *  @author Hassaan Khan (hkahn)
 */

#include <sem.h>


int sem_init( sem_t *sem, int count ){

  if(!sem)
    return -1;

  /* set up the initial reference count and initialize the mutex/cond */
  sem->refcnt=count;
  cond_init(&sem->sem_cv);
  mutex_init(&sem->sem_mut); 

  return 0;
}


int sem_wait( sem_t *sem ){

  if(!sem)
    return -1;

  /* lock the mutex */
  mutex_lock(&sem->sem_mut);

  /* if the reference count is not positive, wait until it is */
  while(sem->refcnt<=0)
    cond_wait(&sem->sem_cv, &sem->sem_mut);
  
  /* decrement the reference count now that it is positive */
  sem->refcnt--;

  /* unlock the mutex */
  mutex_unlock(&sem->sem_mut);
  return 0;   
}


int sem_signal( sem_t *sem ){

  if(!sem)
    return -1;

  /* lock the mutex */
  mutex_lock(&sem->sem_mut);

  /* increase the reference count */
  sem->refcnt++;
  
  /* if the reference count is positive, signal that any
   * of the waiting threads can decrement it
   */
  if(sem->refcnt>0)
    cond_broadcast(&sem->sem_cv);
  
  /* unlock the mutex */
  mutex_unlock(&sem->sem_mut);
  
  return 0;
}


int sem_destroy( sem_t *sem ){
  /* clean up the mutex and condition variable
   * and propagate any errors
   */ 
  if(!sem ||
     mutex_destroy(&sem->sem_mut)<0 ||
     cond_destroy(&sem->sem_cv)<0)
    return -1;
  
  return 0;
}
