/** @file malloc.c
 *  @brief Thread-safe memory allocation.
 *
 *  Provides thread-safe wrappers for basic memory allocation
 *  function by guarding them with a global mutex. If any of
 *  wrappers are called before the thread library is initialized,
 *  they will simply default to calling the original functions
 *  without thread-safety.
 *
 *  @author Lee Salzman (lsalzman)
 */ 

#include <thread.h>
#include <mutex.h>
#include <stdlib.h>

/* whether or not thread-safety has been initialized */
static int initialized = 0;
/* the global lock guarding access to the malloc library */
static mutex_t mallocLock;

/** @brief Initialize the malloc wrappers.
 *
 *  Initializes the global lock and enables thread-safety.
 *
 *  @return Void
 */ 
void
malloc_init (void)
{
	mutex_init (& mallocLock);

	initialized = 1;
}

void *
malloc (size_t size)
{
	void * result;

	/* no thread-safety if not initialized */
	if (! initialized)
	  return _malloc (size);

	/* acquire global lock and call wrapped function */
	mutex_lock (& mallocLock);

	result = _malloc (size);

	mutex_unlock (& mallocLock);

	return result;
}

void
free (void * memory)
{
        /* no thread-safety if not initialized */
	if (! initialized)
	  return _free (memory);
    
	/* acquire global lock and call wrapped function */    
	mutex_lock (& mallocLock);

	_free (memory);

	mutex_unlock (& mallocLock);
}	

void *
realloc (void * memory, size_t size)
{
    void * result;

    /* no thread-safety if not initialized */
    if (! initialized)
      return _realloc (memory, size);

    /* acquire global lock and call wrapped function */
    mutex_lock (& mallocLock);

    result = _realloc (memory, size);

    mutex_unlock (& mallocLock);

    return result;
}

void *
calloc (size_t nmemb, size_t size)
{
    void * result;

    /* no thread-safety if not initialized */
    if (! initialized)
      return _calloc (nmemb, size);

    /* acquire global lock and call wrapped function */
    mutex_lock (& mallocLock);

    result = _calloc (nmemb, size);

    mutex_unlock (& mallocLock);

    return result;
}

