/** @file thread.c
 *  @brief Core thread library implementation.
 *
 *  Implements the core thread management functions.
 *  A simple direct-mapped table is used to map user-land thread ids
 *  (distinct from kernel/system thread ids) to their respective thread
 *  contexts. This table expands as necessary to accomodate new threads.
 *  The distinct thread ids allow for an easy mapping of thread ids to their
 *  contexts as well as allow for the possibility that the kernel recycles
 *  thread ids, so zombie threads do not conflict with the ids of new threads.
 *  A global synchronization resource is used to handle any non-specific
 *  join requests for all threads, with local synchronization resources
 *  on the threads themselves to handle specific join requests for a given
 *  thread. Thread contexts are kept until the threads are actually joined,
 *  but the thread's allocated stack is freed immediately once it exits and
 *  its exit status has been stored in its context. 
 *
 *  @author Lee Salzman (lsalzman)
 */
#include <string.h>
#include <stdlib.h>
#include <thread.h>
#include <syscall.h>
#include <threadlib.h>

/* lock guarding manipulation of thread ids */
static spinlock_t threadsLock = 0;
/* table mapping user-land thread ids to thread contexts */
static thread_context_t ** threads = NULL;
/* the maximum number of thread ids in the table */
static unsigned int threadsCapacity = 0;
/* the size of thread stacks */
static unsigned int threadStackSize = 0;
/* global join resource for threads waiting to join on any other thread */
static thread_resource_t joinResource;

int
thr_init (unsigned int stackSize)
{
	/* make sure the stack size is aligned to a 32 bit boundary */
	if (stackSize % 4)
	  stackSize += 4 - stackSize % 4;

	threadStackSize = stackSize;

	/* initialize the global join resource */
	thread_resource_init (& joinResource);

	/* initialize the memory allocation wrappers */
	malloc_init ();

	if (thr_create(NULL, NULL) < 0)
	  return -1;
	
	return 0;
}

/** @brief Returns context of currently running thread.
 *
 *  Find the thread context matching the system thread id
 *  of the currently running thread.
 *
 *  @return the current thread context, or NULL if no context
 *          has been allocated for the current thread
 */
thread_context_t *
thread_context_current (void)
{
	int userId, systemId;

	/* get the system id of the current thread */	
	systemId = gettid ();

	/* acquire access to the id table */
	spinlock_lock (& threadsLock);

	/* search the id table for a matching context.
	 * thread id 0 is reserved and may not be used.
	 */
	for (userId = 1; userId < threadsCapacity; ++ userId)
	{
		thread_context_t * thread = threads [userId];

		/* check if the context's (if there is one) system id matches */
		if (thread != NULL && systemId == thread -> systemId)
		{
			spinlock_unlock (& threadsLock);
			
			/* found matching context */
			return thread;
		}
	}

	spinlock_unlock (& threadsLock);

	/* no context found */
	return NULL;
}

int
thr_getid (void)
{
	thread_context_t * thread = thread_context_current ();

	/* return the user id of current thread context (if one exists) */
	if (thread != NULL)
	  return thread -> userId;

	/* no context is found only if this is called before the thread library
         * is initialized
         */
	return -1;
}

#include <magic_break.h>
/** @brief Entry-point for new threads.
 *
 *  Calls the thread's entry function and exit with the return status.
 *
 *  @param thread the new thread
 *  @param func the thread's entry function
 *  @param arg the thread's entry argument
 *
 *  @return Void
 */
void
thr_entry (thread_context_t * thread, void * (* func) (void *), void * arg)
{
	/* if the thread entry point happens to get called before thread_fork
	 * returns, the system id has not been placed in the thread yet.
	 * check if the id has been assigned, and if not, then fill it in.
	 * the thread invoking thread_fork will later just overwrite this
	 * with the same value.
	 */
	if (thread -> systemId == -1) 
	  thread -> systemId = gettid ();

	/* call the entry function and exit if it returns */
	thr_exit ((* func) (arg));
}

/** @brief Assign a new user id to a thread.
 *
 *  Search for an empty thread id to assign to a thread context.
 *  If none exists, expand the thread id table to make room for more.
 *
 *  @param thread the thread context to assign an id
 *
 *  @return the new id of the thread, or -1 on failure
 */	 
int
thread_context_allocate_id (thread_context_t * thread)
{
	int userId;

        /* acquire access to the id table */
        spinlock_lock (& threadsLock);

        /* find an empty thread id in the table */
        for (userId = 1; userId < threadsCapacity; ++ userId)
        {
		/* check if the user id has an associated context */
                if (threads [userId] == NULL)
                  break;
        }

	/* if no id was found within the table, expand the table to encompass the overflown id */
        if (userId >= threadsCapacity)
        {
		/* ensure the table size is non-zero, then double it */
		unsigned int newCapacity = (threadsCapacity + 1) * 2;		
                thread_context_t ** newThreads;

		/* allocate new space for the table */
                newThreads = (thread_context_t **) realloc (threads, newCapacity * sizeof (thread_context_t *)); 

		/* error if allocating the new table failed */
                if (newThreads == NULL)
                {
                        spinlock_unlock (& threadsLock);

                        return -1;
                }

		/* null out the new space in the id table */
		memset (newThreads + threadsCapacity, 
                        0, 
			(newCapacity - threadsCapacity) * sizeof (thread_context_t *));

		/* install the grown table */
                threads = newThreads;
                threadsCapacity = newCapacity;
        }

	/* assign the id to the thread */	
        threads [userId] = thread;
        thread -> userId = userId;

        spinlock_unlock (& threadsLock);

	return userId;
}

int
thr_create (void * (* func) (void *), void * arg)
{
	/* allocate the thread context */
	thread_context_t * thread = (thread_context_t *) malloc (sizeof (thread_context_t));
	thread_stack_t * stackBase = NULL;
	char * stack;

	/* if allocation fails, error */
	if (thread == NULL)
	  return -1;

	/* check if the thread has an entry point */
	if (func != NULL)
	{
		/* the thread has an entry point, so a new thread must be created.
		 * allocate a stack for it with some extra space for meta-data and
		 * proper alignment.
		 */
		stack = (char *) malloc (threadStackSize + sizeof (thread_stack_t) + 4);
		if (stack == NULL)
		{
			/* allocation failed, error */
			free (thread);
			return -1;
		}

		/* align the stack to a 32 bit boundry if it is not already */
		if ((unsigned int) stack % 4)
       		   stack += 4 - (unsigned int) stack % 4;
 
		/* point the base of the stack at the top just below the meta-data */
		stackBase = (thread_stack_t *) (stack + threadStackSize - sizeof (thread_stack_t));

		/* fill in the entry function arguments */
		stackBase -> thread = thread;
		stackBase -> entryFunction = func;
		stackBase -> entryArgument = arg;
	}
	else
	/* check if the current thread already has a context */
	if (thread_context_current () != NULL)
	{
		/* the thread has no entry point but already has a context, error */
		free (thread);
		return -1;
	}
	else
	{
		/* the thread has no entry point or context, so assign a context
		 * to the current thread, marking that the thread's original stack is being
		 * reused
		 */
		stack = THREAD_STACK_REUSED;
	}

	/* initialize the synchronization resource for the thread */
	thread_resource_init (& thread -> resource);

	thread -> running = 1;
	thread -> stack = stack;
	thread -> systemId = -1;

	/* assign the thread an id and error if that fails */
	if (thread_context_allocate_id (thread) < 0)
	  goto failure;

	/* if the thread has an entry point, fork a new thread to start at
	 * that entry point, otherwise assign the context to the current thread
	 */ 
	if (func != NULL)
	  thread -> systemId = thread_fork ((void *) thr_entry, stackBase);
	else
	  thread -> systemId = gettid ();

	/* if the thread failed to fork, clear the thread's user id and error */
	if (thread -> systemId < 0)
	{
		spinlock_lock (& threadsLock);
		threads [thread -> userId] = NULL;
		spinlock_unlock (& threadsLock);
		
		goto failure;
	}

	return thread -> userId;

failure:
	/* free the memory allocated for the thread if necessary and error */
	if (stack != THREAD_STACK_REUSED)
	  free (stack);
	free (thread);

	return -1;
}

void
thr_exit (void * status)
{
	thread_context_t * thread = thread_context_current ();
	void * stack = thread -> stack;

	/* acquire access to the current thread's synchronization resource */
	spinlock_lock (& thread -> resource.lock);			

	/* mark the stack as exited by setting its stack to NULL and noting the
	 * the exit status
	 */
	thread -> stack = NULL;
	thread -> exitStatus = status;

	/* check if a thread is joining specifically on this thread */
	if (thread -> resource.waiting != NULL)
	{
		/* mark that the current thread joined with the waiting thread */
		thread -> resource.waiting -> joinId = thread -> userId;

		/* wake up the thread */
		thread_resource_wake (& thread -> resource, 1);
	}		
	else
	{
		/* acquire the global join resource */
		spinlock_lock (& joinResource.lock);

		/* check if any threads are trying to join on all threads */
		if (joinResource.waiting != NULL)
		{
			/* mark that the current thread joined with a waiting thread */
			joinResource.waiting -> joinId = thread -> userId;

			/* wake up the thread */
			thread_resource_wake (& joinResource, 1);
		}

		spinlock_unlock (& joinResource.lock);
	}

	spinlock_unlock (& thread -> resource.lock);

	/* if a stack was allocated for the thread, free it */
	if (stack != THREAD_STACK_REUSED)
	  free (stack);
	
	/* exit out of the system thread */
	exit (0);
}

/** @brief Find any thread to join with.
 *
 *  Searches for any "zombie" threads to join with, and
 *  failing that will wait on the global join resource
 *  for one to become available. The resource of the
 *  joined thread is acquired upon return.
 *
 *  @param joiner the thread attempting a join
 *
 *  @return a thread context joined with the joiner
 */
thread_context_t *
thread_context_join_any (thread_context_t * joiner)
{
        thread_context_t * thread = NULL;
	int userId;

        /* search for any "zombie" threads that are
         * waiting for a join
         */
        for (userId = 1; userId < threadsCapacity; ++ userId)
        {
                thread = threads [userId];
                if (thread == NULL)
                  continue;

                /* acquire the thread's join resource
                 * and choose the thread if it's stack is NULL
                 * (i.e. it exited) and no other thread is waiting
                 * to join on it
                 */
                spinlock_lock (& thread -> resource.lock);
                if (thread -> stack == NULL &&
                    thread -> resource.waiting == NULL)
                  break;

                /* can't join on the thread so release it's resource */
                spinlock_unlock (& thread -> resource.lock);
        }

        /* if there are no "zombie" threads, just wait on the global join
         * resource for some thread to exit
         */
        if (userId >= threadsCapacity)
        {
                /* release the id table */
                spinlock_unlock (& threadsLock);

                /* acquire the join resource and wait on it */
                spinlock_lock (& joinResource.lock);
                thread_resource_sleep (& joinResource, joiner);
                spinlock_unlock (& joinResource.lock);

                /* acquire access to the id table and check the joined thread */
                spinlock_lock (& threadsLock);

                /* acquire access to the joined thread */
                thread = threads [joiner -> joinId];
                spinlock_lock (& thread -> resource.lock);

        }

	return thread;
}
 
int
thr_join (int tid, int * departed, void ** status)
{
	thread_context_t * joiner = thread_context_current (),
			 * thread;

	/* acquire access to the id table */
	spinlock_lock (& threadsLock);

	/* check if the user id is valid */
	if (tid > 0 && tid < threadsCapacity)
	{
		/* if no such thread exists or the thread is attempting to join itself, error */
		thread = threads [tid];
		if (thread == NULL || thread == joiner)
		  return -1;

		/* acquire access to the thread's join resource */
		spinlock_lock (& thread -> resource.lock);
	}
	else
	if (tid == 0)
	{
		/* a global join was requested.
		 * attempt to join with any thread.
		 */
		thread = thread_context_join_any (joiner);
	}
	else
	{
		spinlock_unlock (& threadsLock);

		/* thread id was neither valid nor 0, error */
		return -1;
	}

	/* if the thread's stack is non-NULL (i.e. it is still running)
	 * then sleep on the thread's resource until it exits
	 */
	if (thread -> stack != NULL)
	{
		spinlock_unlock (& threadsLock);

		/* if another thread is already waiting to join, error */
		if (thread -> resource.waiting != NULL)
		{
			spinlock_unlock (& thread -> resource.lock);

			return -1;
		}

		/* wait for the thread to exit */
		thread_resource_sleep (& thread -> resource, joiner);

		spinlock_lock (& threadsLock);
	}

	/* clear the thread's id */
	threads [thread -> userId] = NULL;

	/* release the id table */
	spinlock_unlock (& threadsLock);

	/* relase the thread's join resource */
	spinlock_unlock (& thread -> resource.lock);

	/* return the id of the departed thread if necessary */
	if (departed != NULL)
	  * departed = joiner -> joinId;

	/* return the exit status of the departed thread if necessary */
	if (status != NULL)
	  * status = thread -> exitStatus;

	/* free the thread's context */
	free (thread);

	return 0;
}

