/** @file cond.c
 *  @brief Implementation of condition variables.
 *
 *  Implements condition variables as a simple thread
 *  resource, with waiting threads sleeping on the resource
 *  until a signal wakes them up.
 *
 *  @author Lee Salzman (lsalzman)
 *  @author Hassaan Khan (hkhan)
 */

#include <stdlib.h>
#include <cond.h>

int
cond_init (cond_t * cv)
{
	/* initialize the condition variable's resource */
	thread_resource_init (& cv -> resource);

	return 0;
}

int 
cond_destroy (cond_t * cv)
{
	/* acquire the cond var resource */
	spinlock_lock (& cv -> resource.lock);

	/* if there are threads waiting on the condition
	 * then error
	 */
	if (cv -> resource.waiting != NULL)
	{
		spinlock_unlock (& cv -> resource.lock);

		return -1;
	}

	spinlock_unlock (& cv -> resource.lock);

	return 0;
}


int 
cond_wait (cond_t * cv, mutex_t * mp)
{
	/* acquire the cond var's resource while
	 * the mutex is still locked, so that the sleep
	 * on the resource happens atomically
	 */
	spinlock_lock (& cv -> resource.lock);
  
	/* unlock the mutex since the resource is acquired */
	mutex_unlock (mp);

	/* sleep on the resource until it is signaled or broadcasted */
	thread_resource_sleep (& cv -> resource, thread_context_current ());

	/* release the resource before trying to lock the mutex since it may
	 * cause it to block
	 */
	spinlock_unlock (& cv -> resource.lock);

	/* relock the mutex */
	mutex_lock (mp);

	return 0;
}

int 
cond_signal (cond_t * cv)
{
	/* acquire the cond var's resource */
	spinlock_lock (& cv -> resource.lock);

	/* wake up a thread waiting on the resource, if any */
	thread_resource_wake (& cv -> resource, 1);

	spinlock_unlock (& cv -> resource.lock);

	return 0;
}

int 
cond_broadcast (cond_t * cv)
{
	/* acquire the cond var's resource */
	spinlock_lock (& cv -> resource.lock);

	/* wake up all threads waiting on the resource */
	thread_resource_wake (& cv -> resource, -1);

	spinlock_unlock (& cv -> resource.lock); 
	
	return 0;
}

